"""
---------------------------------------------------------------------------


Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Trajectory Animation Class
"""
from collections.abc import Callable
from typing import Any

import numpy as np
import pandas as pd
import plotly.graph_objects as go
import tqdm
from plotly.subplots import make_subplots

from corl.evaluation.episode_artifact import EpisodeArtifact
from corl.visualization.base_animator import BaseAnimator, FrameAxisBounds
from corl.visualization.platform_plotting_deserializer import PlatformDeserializerPlotlyAnimation


class PlotlyTrajectoryAnimation(BaseAnimator):
    """Produces a trajectory animation given a set of PlatformSteps.
    This can handle both 2D and 3D shapes. The colors of each platform is inferred
    by the name of each platform. For example, when a platform is named "blue0",
    it will be blue. When more than one platform is named with the same color,
    they will all be plotted in the same color. When no color is provided as part
    of the platform name, a random color is chosen.

    """

    def __init__(
        self,
        episode_artifact: EpisodeArtifact,
        platform_serializer_map: dict[str, str],
        progress_bar_func: Callable = tqdm.tqdm,
        platform_color_map: dict[str, str] | None = None,
    ):
        super().__init__(
            episode_artifact=episode_artifact, platform_serializer_map=platform_serializer_map, platform_color_map=platform_color_map
        )

        self.progress_bar_func = progress_bar_func

        self._transition_duration_ms = 10
        self._frame_duration_ms = 30

        self.figure_width = 900
        self.figure_height = 900
        self.frame_axis_bounds: FrameAxisBounds | None = None

    def set_frame_axis_bounds(self, frame_axis_bounds: FrameAxisBounds):
        """Sets the frame axis bounds.

        Parameters
        ----------
        frame_axis_bounds : FrameAxisBounds
            The coordinates of the frame axis
        """
        assert isinstance(frame_axis_bounds, FrameAxisBounds)
        if self.get_dimensionality() == 2:
            assert (
                frame_axis_bounds.zbounds is None
            ), """
            Dimension mismatch between the world and the frame axis bounds.
            z-coordinates should be None for a 2-dimensional world."""

        self.frame_axis_bounds = frame_axis_bounds

    @property
    def platform_deserializer(self):
        """Platform deserializer as prop"""
        deserializer = self.get_deserializer()
        assert issubclass(
            deserializer, PlatformDeserializerPlotlyAnimation
        ), f"Deserializer is expected to subclass: {PlatformDeserializerPlotlyAnimation.__name__}, deserializer: {deserializer} is invalid"
        return deserializer()

    def get_metadata_names(self):
        """Returns a list of the metadata that is available to display as text
        above the platform"""
        metadata_keys = []
        # This is a list of serialized platform data for each platform for one step
        for platform_serialized_data in self.config.episode_artifact.steps[0].platforms:
            metadata = self.platform_deserializer.get_metadata_for_text_display(platform_serialized_data)
            metadata_keys.extend(list(metadata.keys()))
        metadata_keys = list(set(metadata_keys))
        metadata_keys.sort()
        return metadata_keys

    def get_dimensionality(self):
        """Returns the expected dimensionality of the world."""
        return self.platform_deserializer.get_dimensionality(serialized_data=self.config.episode_artifact.steps[0].platforms[0])

    @staticmethod
    def _check_and_parse_position(position: list | tuple, position_units: list | tuple):
        assert len(position) in [
            1,
            2,
            3,
        ], f"""
        The position must either be 1- 2- or 3- dimensional.
        Found: {len(position)} dimensions in {position}"""
        coordinate_keys = ["x", "y", "z"]
        parsed_position = {}
        for idx, val in enumerate(position):
            key_name = f"{coordinate_keys[idx]}[{position_units[idx]}]"
            parsed_position[key_name] = val
        return parsed_position

    def generate_animation(  # noqa: PLR0915
        self,
        selected_metadata: list[str] | None = None,
    ) -> tuple[go.Figure, list[go.Figure], dict[str, pd.DataFrame]]:
        """Generates the animation

        Parameters
        ----------
        selected_metadata : List[str]
            A list of selected metadata to display above each platform

        Returns
        -------
        go.Figure
            The figure as an animation
        List[go.Figure]
            A list of the frames that make up the animation
        Dict[str, pd.DataFrame]
            A table of the underlying data for each platform.
        """

        # #####################################################################
        # Configure Plotly Figures
        # #####################################################################
        expected_dims = self.get_dimensionality()

        # Redraw is False for 2D - this speeds up the animation
        redraw = expected_dims != 2 or 1
        trajectory_position = {"row": 1, "col": 1}
        subplot_kwargs = {"rows": 1, "cols": 1, "specs": [[{"type": "scatter" if expected_dims <= 2 else "scene"}]]}
        fig = make_subplots(**subplot_kwargs)
        fig.update_layout({"width": self.figure_width, "height": self.figure_height})
        # When region bounds are passed in, constrain the view of the animation to the region bounds.
        layout_kwargs = self._create_layout_with_region_bounds()
        if layout_kwargs:
            fig.update_layout(**layout_kwargs)

        # Hover Menu
        fig.update_layout({"hoverlabel": {"bgcolor": "white", "font": {"color": "black", "size": 12}}, "hovermode": "closest"})

        fig = self.create_play_and_pause_buttons(fig, redraw=redraw)
        # Time slider: "steps" are populated below
        sliders_dict = self.initial_sliders_dict()
        stored_figures = []
        frames = []

        # #####################################################################
        # Deserialize platform steps & Plot
        # #####################################################################
        # Store the platform steps to construct a table to return
        platform_steps: dict = {}
        # The step number is the step within the simulation. This starts at step 1
        # and excludes the initial state (e.g. step 0). The `idx` or `step_idx`
        # corresponds to the index of the platform steps. These two values are
        # off by 1.
        step_num = 1
        for step_data in self.progress_bar_func(self.config.episode_artifact.steps):
            step_idx = step_num - 1
            subplot = make_subplots(**subplot_kwargs)

            # Steps start at 1 because the initial state is not stored
            step_blob = {}
            for serialized_platform_data in step_data.platforms:
                position = self.platform_deserializer.get_position(serialized_platform_data)
                position_units = self.platform_deserializer.get_position_units(serialized_platform_data)
                platform_name = self.platform_deserializer.get_platform_name(serialized_platform_data)
                metadata = self.platform_deserializer.get_metadata_for_text_display(serialized_platform_data)
                filtered_metadata = {}
                if selected_metadata:
                    filtered_metadata = {key: value for key, value in metadata.items() if key in selected_metadata}
                step_blob = self._check_and_parse_position(position, position_units)
                step_blob["step"] = step_num
                step_blob.update(metadata)

                if platform_name in platform_steps:
                    platform_steps[platform_name].append(step_blob)
                else:
                    platform_steps[platform_name] = [step_blob]

                position_units_as_string = ",".join(position_units)
                position_as_string = ",".join([str(round(pos, 2)) for pos in position])
                hover_template = f"""
                    <b>Platform: {platform_name}</b><br>
                    Position [{position_units_as_string}]: ({position_as_string})<br>
                    """

                # Draws a point representative of each platform
                platform_point = draw_point(
                    input_coordinates=tuple(position),
                    marker_color=self.config.platform_color_map[platform_name] if self.config.platform_color_map else "black",
                    plot_text=generate_overhead_text(filtered_metadata),
                    text_position="top center",
                    hover_template=hover_template,
                    name=platform_name,
                )
                subplot.add_trace(platform_point, **trajectory_position)
                # Initialize the first frame, which is stored separately from the subplots
                if step_num == 1:
                    fig.add_trace(trace=platform_point, **trajectory_position)

            custom_polygons = self._draw_custom_polygons(step_idx)
            if custom_polygons is not None:
                for polygon in custom_polygons:
                    if step_num == 1:
                        fig.add_trace(trace=polygon, **trajectory_position)
                    subplot.add_trace(trace=polygon, **trajectory_position)

            subplot.update_layout(**layout_kwargs)

            stored_figures.append(subplot)
            # NOTE: the name of the frame and the slider step must match
            frames.append(go.Frame(name=step_num, data=subplot.data, layout=subplot.layout))
            # Add the slider step
            sliders_dict["steps"].append(self.one_slider_step_animation(step=step_num, redraw=redraw))
            step_num += 1
        platform_step_tables = {}
        for platform_name, platform_data in platform_steps.items():
            platform_step_tables[platform_name] = pd.DataFrame(platform_data)
        fig["layout"]["sliders"] = [sliders_dict]
        fig["frames"] = frames
        return fig, stored_figures, platform_step_tables

    def _draw_custom_polygons(self, _) -> list[go.Scatter | go.Scatter3d] | None:  # noqa: PLR6301
        """This is an optional method to add additional shapes to
        each frame of the animation. Override this method to generate
        custom polygons to plot. There is a helper function `draw_custom_polygon`
        that may help with this."""
        return None

    def _create_layout_with_region_bounds(self):
        """Generates the layout keyword arguments for Figure"""
        layout_kwargs = {}
        if self.frame_axis_bounds:
            # #################################################################
            # Layout for 2D ScatterPlot
            # #################################################################
            layout_kwargs["xaxis"] = {"range": self.frame_axis_bounds.xbounds, "autorange": False}
            layout_kwargs["xaxis_title"] = "x-Position"
            # The scaleanchor sets the range of the y-axis to change with the range of the x-axis
            # such that the scale of pixels per unit is in a constant ratio. Both axes are still zoomable
            # but when you zoom into one, the other will zoom the same amount, keeping a fixed midpoint.
            # The scaleratio determines the pixel to unit scale ratio (e.g. 1:1 with x-axis)
            layout_kwargs["yaxis"] = {"range": self.frame_axis_bounds.ybounds, "autorange": False, "scaleanchor": "x", "scaleratio": 1}
            layout_kwargs["yaxis_title"] = "y-Position"
            # #################################################################
            # Layout for 3D ScatterPlot - "Scene" layout
            # #################################################################
            if self.frame_axis_bounds.zbounds:
                # These options are not available for the 3D plot layouts
                del layout_kwargs["yaxis"]["scaleanchor"]
                del layout_kwargs["yaxis"]["scaleratio"]
                layout_kwargs = {"scene": layout_kwargs}
                layout_kwargs["scene"]["zaxis"] = {"range": self.frame_axis_bounds.ybounds, "autorange": False}
                layout_kwargs["scene"]["zaxis_title"] = "z-Position"
                # Locks the aspect ratio so that the animation doesn't change between frames.
                layout_kwargs["scene"]["aspectratio"] = {"x": 1, "y": 1, "z": 1}
        return layout_kwargs

    def one_slider_step_animation(self, step: Any, redraw=False):
        """Returns a dictionary for one slider step for animations.
        This is one element in the `sliders_dict["steps"]` list.

        Parameters
        ----------
        step : Any
            The attribute in the data that defines a step.
        frame_duration_ms : int, optional
            The time in milliseconds of each frame, by default 30
        transition_duration_ms : int, optional
            The time in milliseconds of the transition of each frame, by default 30

        Returns
        -------
        dict
            A dict for one slider step.
        """
        return {
            "args": [
                [step],
                {
                    "frame": {"duration": self._frame_duration_ms, "redraw": redraw},
                    "mode": "immediate",
                    "transition": {"duration": self._transition_duration_ms},
                },
            ],
            "label": step,
            "method": "animate",
        }

    def create_play_and_pause_buttons(self, fig: go.Figure, redraw=False):
        """Updates a plotly `fig_dict` with the parameters for creating play and pause buttons

        Parameters
        ----------
        fig : go.Figure
            Figure option

        Returns
        -------
        Dict:
            The original fig_dict with the `fig_dict["layout"]["updatemenus"]` property
            updated to include buttons.
        """

        # Append the play and pause buttons to it
        update_menus = []
        update_menus.append(
            {
                "buttons": [
                    {
                        "args": [
                            None,
                            {
                                "frame": {"duration": self._frame_duration_ms, "redraw": redraw},
                                "fromcurrent": True,
                                "transition": self._transition_duration_ms,
                            },
                        ],
                        "label": "Play",
                        "method": "animate",
                    },
                    {
                        "args": [[None], {"frame": {"duration": 0, "redraw": redraw}, "mode": "immediate", "transition": {"duration": 0}}],
                        "label": "Pause",
                        "method": "animate",
                    },
                ],  # Button Placement
                "direction": "left",
                "pad": {"r": 10, "t": 87},
                "showactive": False,
                "type": "buttons",
                "x": 0.1,
                "xanchor": "right",
                "y": 0,
                "yanchor": "top",
            }
        )
        fig.update_layout(updatemenus=update_menus)
        return fig

    def initial_sliders_dict(self, step_descriptor: str = "Step"):
        """Initializes the properties for a slider bar. The `step` key must still be filled out

        Parameters
        ----------
        transition_duration_ms : int
            The transition time in seconds for each frame.
        step_descriptor : str, optional
            A string that describes each step (years, step, etc.), by default "Step"

        Returns
        -------
        dict
            A dictionary with the slider properties initializes.
        """
        return {
            "active": 0,
            "yanchor": "top",
            "xanchor": "left",
            "currentvalue": {"font": {"size": 12}, "prefix": f"{step_descriptor}: ", "visible": True, "xanchor": "right"},
            "transition": {"duration": self._transition_duration_ms},
            "pad": {"b": 10, "t": 50},
            "len": 0.9,
            "x": 0.1,
            "y": 0,
            "steps": [],
        }


def draw_custom_polygon(
    x_coordinates: np.ndarray | (list | tuple),
    y_coordinates: np.ndarray | (list | tuple),
    z_coordinates: np.ndarray | (list | tuple) | None = None,
    color: str = "black",
    fill: bool = True,
    name: str | None = None,
) -> go.Scatter | go.Scatter3d:
    """Draws a polygon according to the coordinates provided.

    Parameters
    ----------
    x_coordinates : Union[np.ndarray, List, Tuple]
        An array of x-coordinates that define the polygon
    y_coordinates : Union[np.ndarray, List, Tuple]
        An array of y-coordinates that define the polygon
    color : str, optional
        Color of the shape, by default "black"
    fill : bool, optional
        When True, fills the shape , by default True
    name : Union[str, None], optional
        The name of the trace, will show in the hover menu and legend, by default None

    Returns
    -------
    plotly.graph_objs.Scatter
        Returns a plotly scatter graph object

    Examples
    --------
    >>> wedge_array = matplotlib_wedge_as_array(center=(0,0), radius=10, start_angle_deg=0, stop_angle_deg=10)
    >>> wedge = draw_custom_polygon(x_coordinates=wedge_array[:,0], y_coordinates=wedge_array[:,1], name="wedge")
    >>> fig = go.Figure(wedge)
    >>> fig.show()
    """
    fill_value = "toself" if fill else None
    if z_coordinates:
        polygon = go.Scatter3d(
            x=x_coordinates, y=y_coordinates, z=z_coordinates, line=dict(color=color), fill=fill_value, name=name, mode="lines"
        )
    else:
        polygon = go.Scatter(x=x_coordinates, y=y_coordinates, line=dict(color=color), fill=fill_value, name=name, mode="lines")

    return polygon


def draw_point(
    input_coordinates: tuple,
    marker_color: str = "black",
    marker_size: int | None = None,
    plot_text: str | None = None,
    hover_template: str | None = None,
    name: str | None = None,
    text_position: str | None = "top center",
) -> go.Scatter | go.Scatter3d:
    """Draws a dot.

    Parameters
    ----------
    input_coordinates : Tuple
        The position of the dot (2D) or sphere (3D).
    marker_color : str, optional
        The color of the marker, by default "black"
    plot_text : Union[str, None], optional
        The text that will go above the the dot, by default None
    hover_template : Union[str, None], optional
        The hover template., by default None
    name : Union[str, None], optional
        Name for legend, by default None
    text_position : Union[str, None], optional
        Position of text, by default "top center"

    Returns
    -------
    plotly.graph_objs.Scatter
        Returns a plotly scatter graph object

    Examples
    --------
    >>> dot = draw_point(xy=(10,10), plot_text="a dot")
    >>> fig = go.Figure(dot)
    >>> fig.show()
    """
    assert len(input_coordinates) in [1, 2, 3], "Input coordinates must either be 1- 2- or 3-dimensions"
    mode = "markers"

    if plot_text:
        mode = f"{mode}+text"
    if len(input_coordinates) < 3:
        dot = go.Scatter(
            x=[input_coordinates[0]],
            y=[0] if len(input_coordinates) == 1 else [input_coordinates[1]],
            marker={"color": marker_color, "size": marker_size},
            mode=mode,
            name=name,
            text=plot_text,
            textposition=text_position,
            hovertemplate=hover_template,
        )
    else:
        dot = go.Scatter3d(
            x=[input_coordinates[0]],
            y=[input_coordinates[1]],
            z=[input_coordinates[2]],
            marker={"color": marker_color, "size": marker_size},
            mode=mode,
            name=name,
            text=plot_text,
            textposition=text_position,
            hovertemplate=hover_template,
        )

    return dot


def generate_overhead_text(metadata: dict[str, int | float]) -> str | None:
    """Generates a block of text, separated by new lines. Numeric values will
    be rounded (4-digits)

    Parameters
    ----------
    data : Dict
    Returns
    -------
    str
        Generated block of text, joined together by a newline.
    """
    if len(metadata) == 0:
        return None

    text_block = []
    for name, value in metadata.items():
        if isinstance(value, str | int):
            text_line = f"{name}: {value}"
            text_block.append(text_line)
        elif isinstance(value, float):
            text_line = f"{name}: {value: .4f}"
            text_block.append(text_line)

    # Only return the text block as a string if text_block is not empty
    return "<br>".join(text_block)



import numpy as np
from plotly.graph_objs import Scatter, Scatter3d

from corl.dones.docking_1d.dones import DockingDoneFunction
from corl.visualization.plotly_animator import PlotlyTrajectoryAnimation, draw_custom_polygon


class Docking1dAnimation(PlotlyTrajectoryAnimation):
    def _draw_custom_polygons(self, _) -> list[Scatter | Scatter3d] | None:
        """
        Adds the docking region and maximum velocity threshold for successful docking to the trajectory animation
        """
        agents = self.config.episode_artifact.env_config["agents"]
        custom_polygons = []
        for agent_id, agent_name in self.config.episode_artifact.agent_to_platforms.items():
            agent_config = agents[agent_id].class_config.config
            done_config = [(done["config"]) for done in agent_config["dones"] if done["name"] == DockingDoneFunction.__name__]
            docking_region_radius = done_config[0]["docking_region_radius"]["value"]
            done_config[0]["docking_region_radius"]["units"]

            velocity_threshold = done_config[0]["velocity_threshold"]["value"]
            velocity_units = done_config[0]["velocity_threshold"]["units"]

            x_coordinates = []
            y_coordinates = []
            t = 0.0
            while t <= 2 * np.pi:
                x_coordinates.append(docking_region_radius * np.cos(t))
                y_coordinates.append(docking_region_radius * np.sin(t))
                t += 0.1
            custom_polygons.append(
                draw_custom_polygon(
                    x_coordinates=x_coordinates,
                    y_coordinates=y_coordinates,
                    fill=False,
                    color=self.config.platform_color_map[agent_name[0]] if self.config.platform_color_map else "black",
                    name=f"Docking Region - velocity threshold: {velocity_threshold} [{velocity_units}]",
                )
            )
        return custom_polygons



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from __future__ import annotations

import dataclasses
import typing

from corl.evaluation.metrics.metric import TerminalMetric


@dataclasses.dataclass
class Rate(TerminalMetric):
    """Terminal Metric to represent the number of occurrences of an event vs total"""

    occurrence: int
    total: int

    def __add__(self, rhs: typing.Any) -> Rate:
        if isinstance(rhs, Rate):
            return Rate(self.occurrence + rhs.occurrence, self.total + rhs.total)
        raise NotImplementedError()

    def __truediv__(self, denominator: typing.Any):
        raise RuntimeError("Can not divide a Rate metric")

    def __lt__(self, rhs: typing.Any) -> bool:
        if isinstance(rhs, Rate):
            raise NotImplementedError()
        if isinstance(rhs, int):
            return self.occurrence / self.total < rhs / self.total
        if isinstance(rhs, float):
            return self.occurrence / self.total < float(rhs)
        raise RuntimeError(f"Unknown how to compute lt operator between Rate and {type(rhs)}")

    def __gt__(self, rhs: typing.Any) -> bool:
        if isinstance(rhs, Rate):
            raise NotImplementedError()
        if isinstance(rhs, int):
            return self.occurrence / self.total > rhs / self.total
        if isinstance(rhs, float):
            return self.occurrence / self.total > rhs
        raise RuntimeError(f"Unknown how to compute lt operator between Rate and {type(rhs)}")

    def __sub__(self, rhs: typing.Any) -> Rate:
        if isinstance(rhs, Rate):
            raise NotImplementedError()
        if isinstance(rhs, int):
            return Rate(self.occurrence - rhs, self.total)
        if isinstance(rhs, float):
            return Rate(int((self.occurrence - rhs) * self.total), self.total)
        raise RuntimeError(f"Unknown how to subtract between Rate and {type(rhs)}")



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Logs extra environment metrics specific to Pong.
"""
from ray.rllib import BaseEnv
from ray.rllib.algorithms.callbacks import DefaultCallbacks
from ray.rllib.evaluation.episode_v2 import EpisodeV2


class CustomPongMetrics(DefaultCallbacks):
    """Logs the number of ball hits for each paddle"""

    def on_episode_end(self, *, base_env: BaseEnv, episode: EpisodeV2, **_) -> None:  # noqa: PLR6301
        """At the end of each episode, log the total paddle hits"""
        if isinstance(episode, Exception):
            return

        env = base_env.get_sub_environments()[episode.env_id]
        # Cycle through the platforms and log the ball hit for each
        for platform_name, simulator in env.state.sim_platforms.items():
            key = f"{platform_name}/ball_hits"
            episode.custom_metrics[key] = simulator.paddle.ball_hits

            paddle_type = simulator.paddle_type.name
            key = f"{paddle_type}/ball_hits"
            episode.custom_metrics[key] = simulator.paddle.ball_hits



"""
--------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------

Wrapper glue which returns the magnitude of its wrapped glue as an observation.

Author: Jamie Cunningham
"""
from collections import OrderedDict
from functools import cached_property

import numpy as np

from corl.glues.base_wrapper import BaseWrapperGlue, BaseWrapperGlueValidator
from corl.glues.common.observe_sensor import ObserveSensor
from corl.libraries.property import BoxProp, DictProp
from corl.libraries.units import Quantity, corl_quantity


class MagnitudeGlueValidator(BaseWrapperGlueValidator):
    """
    wrapped_glue: ObserveSensor
        wrapped observe sensor
    """

    wrapped: ObserveSensor


class MagnitudeGlue(BaseWrapperGlue):
    """
    Returns the magnitude of the wrapped glue's returned observation.
    """

    class Fields:
        """
        field data
        """

        MAG = "mag"

    def __init__(self, **kwargs) -> None:
        self.config: MagnitudeGlueValidator
        super().__init__(**kwargs)

        self._uname = f"{self.glue().get_unique_name()}_Magnitude"

        wrapped_prop = self.glue().observation_prop
        if not wrapped_prop or not isinstance(wrapped_prop, DictProp):
            raise RecursionError(
                "the glue the magnitude glue is wrapping does not have a valid obs space"
                "This glue expects the wrapped glue to provide a obs space and that space"
                " must be a Dict space"
            )

        child_prop = wrapped_prop.spaces[self.config.wrapped.Fields.DIRECT_OBSERVATION]
        assert isinstance(child_prop, BoxProp)
        self._prop = DictProp(
            spaces={
                self.Fields.MAG: BoxProp(
                    low=[0.0],
                    high=[np.maximum(np.linalg.norm(child_prop.low), np.linalg.norm(child_prop.high))],
                    shape=(1,),
                    dtype=np.dtype("float32"),
                    unit="dimensionless",
                )
            }
        )

    @staticmethod
    def get_validator() -> type[MagnitudeGlueValidator]:
        """returns the validator for this class

        Returns:
            MagnitudeGlueValidator -- A pydantic validator to be used to validate kwargs
        """
        return MagnitudeGlueValidator

    def get_unique_name(self) -> str:
        return self._uname

    @cached_property
    def observation_prop(self):
        return self._prop

    def get_observation(self, other_obs: OrderedDict, obs_space, obs_units):
        obs: Quantity = self.glue().get_observation(other_obs, obs_space, obs_units)  # type: ignore
        if isinstance(obs, dict):
            obs = obs[self.config.wrapped.Fields.DIRECT_OBSERVATION]
        else:
            raise RuntimeError("This glue is not capable of handing wrapped obs that are not a dict")
        d = OrderedDict()
        d[self.Fields.MAG] = corl_quantity()(np.array([np.linalg.norm(obs.m)], dtype=np.float32), "dimensionless")
        return d



"""
---------------------------------------------------------------------------


Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Functions for agent steps section of the stremalit app.
"""

import pandas as pd
import plotly.express as px
import plotly.graph_objs as go
import streamlit as st
from streamlit.runtime.state.session_state_proxy import SessionStateProxy

from corl.visualization.episode_artifact_to_tables import FlattenedAgentStep
from corl.visualization.streamlit_app.utils import checkbox_grid, confidence_interval, st_dataframe_pagination, ui_dataframe_filter


def agent_steps_section(  # noqa: PLR0915
    tables_dict: dict[str, pd.DataFrame],
    sess_state: SessionStateProxy,
) -> SessionStateProxy:
    """
    Creates the streamlit Agent Steps section

    Parameters
    ----------
    tables_dict : Dict[str, pd.DataFrame]
        This is expected to be EpisodeArtifactTables.tables
    sess_state : SessionStateProxy
        The streamlit session state
    obs_units_flattened : Dict[str, str]
        A dictionary where the key corresponds to the attribute name and the
        value corresponds to the unit (str)
    """

    _section_headers()
    sess_state.show_mean_with_ci = False
    agent_step_table = tables_dict["FlattenedAgentStep"]

    ###########################################################################
    # User Input: Select Attributes to Plot
    ###########################################################################
    user_selected_attributes = expander_select_attributes(agent_step_table)

    test_case_idx = st.number_input(
        label="Test Case Number", min_value=agent_step_table["test_case"].min(), max_value=agent_step_table["test_case"].max()
    )

    sess_state.test_case_selected = test_case_idx
    attributes_filter = agent_step_table["attribute_name"].isin(user_selected_attributes)
    # Depending on the selections below, the test case filter may not be used.
    agents = agent_step_table["agent_id"].unique().tolist()
    agents.sort()

    # #####################################################
    # User selections for how to display plots: Radio
    # buttons
    # #####################################################
    # Options for displaying on the same / separate plots
    # and options to show all trajectories (raw) and by
    # aggregate.
    options_dict = {
        "same_plot": "Display traces on the same plot",
        "different_plots": "Display traces on separate plots",
        "show_all": "Show all test cases",
        "aggregate": "Aggregate across test cases (mean +/- 95% CI)",
    }
    plot_layout = (
        st.radio(label="radio", label_visibility="hidden", horizontal=True, options=list(options_dict.values()))
        or options_dict["same_plot"]
    )

    sess_state.show_mean_with_ci = plot_layout == options_dict["aggregate"]
    if not hasattr(sess_state, "custom_y_axes_dict"):
        sess_state.custom_y_axes_dict = {}
        sess_state.custom_y_axes_dict = {_agent_id: {} for _agent_id in agents}

    # #####################################################################
    # Displays different agents in different tabs
    # #####################################################################
    agent_tabs = st.tabs(agents)
    for idx, agent_id in enumerate(agents):
        # Filter by agent and selected attributes
        agent_filter = agent_step_table["agent_id"] == agent_id
        # Makes a copy so that we do not modify the underlying dataframe
        _subset_df = agent_step_table.loc[attributes_filter & agent_filter].copy()
        if _subset_df.empty:
            continue
        # Creates agent tabs
        with agent_tabs[idx]:
            # #####################################################################
            # Displays Plots & Data and nested tab for each agent
            # #####################################################################
            plot_and_data_tabs = st.tabs(["Plots", "Tabular Data"])
            # #################################################################
            # Nested Tab 0: Plotly Figure & Options
            # #################################################################
            with plot_and_data_tabs[0]:
                # #############################################################
                # User Selections for how to display plots: Checkbox options
                # #############################################################
                _subset_df, plot_kwargs = _expander_checkboxes_xaxes_options_and_filter(_subset_df)

                attributes_in_dataframe = _subset_df["attribute_name"].unique().tolist()
                attributes_in_dataframe.sort()

                # #############################################################
                # Y-axis View Options
                # #############################################################
                _expander_set_yaxes_view_options(
                    agent_steps_table=_subset_df,
                    agent_id=agent_id,
                    plot_layout=plot_layout,
                    options_dict=options_dict,
                    sess_state=sess_state,
                    attributes_in_dataframe=attributes_in_dataframe,
                )

                # #############################################################
                # Filters dataframe when necessary & builds the plotly kwargs
                # #############################################################
                test_case_filter = _subset_df["test_case"] == test_case_idx

                # Aggregate with mean and 95% CI option
                # When shown by aggregate, each attributed selected must be on a separate plot
                # for the 95% confidence intervals to render correctly.
                if plot_layout == options_dict["aggregate"]:
                    _subset_df = _expander_filter_dataframe(tables_dict=tables_dict, filtered_agent_table=_subset_df, sess_state=sess_state)
                    _subset_df, plot_kwargs = _plot_layout_aggregate_option(agent_steps_table=_subset_df, plot_kwargs=plot_kwargs)
                    sess_state.show_mean_with_ci = True

                # Display on separate plots
                if plot_layout == options_dict["different_plots"]:
                    plot_kwargs = _plot_layout_different_plots_option(agent_steps_table=_subset_df, plot_kwargs=plot_kwargs)
                    _subset_df = _subset_df.loc[test_case_filter]

                # Display on the same plot
                if plot_layout == options_dict["same_plot"]:
                    plot_kwargs = _plot_layout_same_plot_option(plot_kwargs=plot_kwargs)
                    _subset_df = _subset_df.loc[test_case_filter]

                # Display all test cases: when all traces across test cases are shown,
                # each attribute gets its own plot
                if plot_layout == options_dict["show_all"]:
                    _subset_df = _expander_filter_dataframe(tables_dict=tables_dict, filtered_agent_table=_subset_df, sess_state=sess_state)
                    plot_kwargs = _plot_layout_show_all_option(agent_steps_table=_subset_df, plot_kwargs=plot_kwargs)

                # #####################################################################
                # Creates the Plotly Figure
                # #####################################################################
                category_orders = {"attribute_name": attributes_in_dataframe}
                # When the test case column is present, order by test case
                if "test_case" in _subset_df.columns:
                    test_cases = _subset_df["test_case"].unique().tolist()
                    test_cases.sort()
                    category_orders.update({"test_case": test_cases})

                fig = px.line(_subset_df, markers=True, category_orders=category_orders, hover_data=_subset_df.columns, **plot_kwargs)
                if sess_state.get("show_mean_with_ci"):
                    fig = update_figure_with_confidence_intervals(agent_steps_table=_subset_df, fig=fig, plot_kwargs=plot_kwargs)

                if hasattr(sess_state, "custom_y_axes_dict") and sess_state.custom_y_axes_dict[agent_id]:
                    fig = update_figure_with_y_axes_view_options(fig=fig, custom_y_axes_dict=sess_state.custom_y_axes_dict[agent_id])

                # #############################################################
                # Display the Figure on streamlit
                # #############################################################
                # Do not sync the y-axes for faceted plots
                fig.update_yaxes(matches=None)
                fig.update_yaxes({"tickfont": {"size": 20}})
                fig.update_xaxes({"tickfont": {"size": 20}, "titlefont": {"size": 20}})
                st.plotly_chart(fig, use_container_width=True)

            # #################################################################
            # Nested Tab 2: Tabular data for underlying Plotly Figure
            # #################################################################
            with plot_and_data_tabs[1]:
                # Save the state of the page so that it doesn't restart upon every call
                sess_state.pages_dict = st_dataframe_pagination(
                    dataframe=_subset_df, table_key="FlattenedAgentStep", pages_dict=sess_state.get("pages_dict", {})
                )
                st.markdown(
                    f"""**Table**: Displaying a subset of the agents steps table.
                    The field `attribute_name` is limited to: ```{attributes_in_dataframe}```"""
                )
    return sess_state


def expander_select_attributes(agent_step_table: pd.DataFrame) -> list[str]:
    """Creates a section grouped by attribute type (e.g. observations, actions, rewards) for a grid
    of attributes to select from. Returns the selected attributes.

    Parameters
    ----------
    agent_step_table : pd.DataFrame
        The agent steps table from EpisodeArtifactTable
    obs_units_flattened : dict
        A dictionary where the key corresponds to the attribute name (str)
        and the value corresponds to the units (str)

    Returns
    -------
    List[str]
        A list of selected attributes
    """
    selections_table = agent_step_table[["attribute_descrip", "attribute_name", "units"]].drop_duplicates()
    idx_has_units = selections_table["units"].notnull()
    selections_table["attribute_name_w_units"] = selections_table["attribute_name"].copy()
    selections_table.loc[idx_has_units, "attribute_name_w_units"] = (
        selections_table.loc[idx_has_units, "attribute_name"] + "[" + selections_table.loc[idx_has_units, "units"] + "]"
    )
    user_selected_attributes = []

    with st.expander("**Select Observations to Plot**", expanded=True):
        st.markdown("**Observations**")
        possible_selections = selections_table.loc[
            selections_table["attribute_descrip"] == FlattenedAgentStep.AttributeDescription.observation
        ]["attribute_name"].tolist()

        possible_selections_with_units = selections_table.loc[
            selections_table["attribute_descrip"] == FlattenedAgentStep.AttributeDescription.observation
        ]["attribute_name_w_units"].tolist()

        # # NOTE: We append units to the display of possible units, but this is not how they are stored within the table.
        user_selected_attributes.extend(
            checkbox_grid(
                possible_choices=possible_selections_with_units,
                key_prefix="obs_corl_eval",
                possible_choices_return_value=possible_selections,
            )
        )

    with st.expander("**Select Actions to Plot**", expanded=True):
        st.markdown("**Actions**")
        possible_selections = selections_table.loc[selections_table["attribute_descrip"] == FlattenedAgentStep.AttributeDescription.action][
            "attribute_name"
        ].tolist()
        # NOTE: We append units to the display of possible units;
        # this is not how they are stored within the table so we need to keep
        # track of it separately.
        possible_selections_with_units = selections_table.loc[
            selections_table["attribute_descrip"] == FlattenedAgentStep.AttributeDescription.action
        ]["attribute_name_w_units"].tolist()

        user_selected_attributes.extend(
            checkbox_grid(
                possible_choices=possible_selections_with_units,
                key_prefix="action_corl_eval",
                possible_choices_return_value=possible_selections,
            )
        )

    with st.expander("**Select Rewards to Plot**", expanded=True):
        st.markdown("**Rewards**")
        possible_selections = selections_table.loc[
            selections_table["attribute_descrip"].isin(
                [FlattenedAgentStep.AttributeDescription.reward, FlattenedAgentStep.AttributeDescription.cumulative_reward]
            )
        ]["attribute_name"].tolist()
        user_selected_attributes.extend(checkbox_grid(possible_choices=possible_selections, key_prefix="reward_corl_eval"))

    return user_selected_attributes


def append_confidence_interval_columns(
    dataframe: pd.DataFrame,
    steps_col: str = "steps",
    attribute_name_col: str = "attribute_name",
    attribute_value_col: str = "attribute_value",
    conf_interval: int = 95,
):
    """Creates two columns indicating the upper and lower bounds of the
    confidence interval, where which confidnece interval is defined by
    `conf_interval`.

    Groups the `dataframe` by `steps_col` and `attribute_name` and applies
    the function `utils.confidence_interval` to compute the confidence
    intervals.

    Parameters
    ----------
    dataframe : pd.DataFrame
        A dataframe with at least the FlattenedAgentSteps schema.
        That is, this should be a trajectory.
    steps_col : str, optional
        The column name where the steps are stored, by default 'steps'
    attribute_name_col : str, optional
        The name of the column that stores the attribute names, by default 'attribute_name'
    attribute_value_col : str, optional
        The name of the column that stores the attribute values, by default 'attribute_value'
    conf_interval : int, optional
        An number indicating the confidence interval, by default 95.

    Returns
    -------
    pd.DataFrame:
        A dataframe with columns: `steps_col`, `attribute_name_col`,
        `<attribute_value_col>_mean`, `<attribute_value_col>_ci_lower`, `<attribute_value_col>_ci_upper`
    """

    dataframe = dataframe.copy()
    dataframe[f"{attribute_value_col}_mean"] = dataframe[attribute_value_col].copy()
    dataframe[f"{attribute_value_col}_ci_lower"] = dataframe[attribute_value_col].copy()
    dataframe[f"{attribute_value_col}_ci_upper"] = dataframe[attribute_value_col].copy()
    return dataframe.groupby([steps_col, attribute_name_col], as_index=False).agg(
        {
            f"{attribute_value_col}_mean": "mean",
            f"{attribute_value_col}_ci_lower": lambda column_values: confidence_interval(column_values, which=conf_interval)[0],
            f"{attribute_value_col}_ci_upper": lambda column_values: confidence_interval(column_values, which=conf_interval)[1],
        }
    )


def update_figure_with_confidence_intervals(agent_steps_table: pd.DataFrame, fig: go.Figure, plot_kwargs: dict) -> go.Figure:
    """Updates the plotly figure with the custom 95% confidence interval view

    Parameters
    ----------
    agent_steps_table : pd.DataFrame
        The agent steps table from EpisodeArtifactTable or a subset of it.
    fig : go.Figure
        The plotly figure object with the mean already plotted.
    plot_kwargs : Dict
        plotly kwargs so far.

    Returns
    -------
    go.Figure
        Updated figure.
    """

    # The order the plots appear may not match how its internally stored and referenced by
    # the column / row within plotly. We grab the layout object and check the name of the title
    # to ensure the 95% CI are plotted on the correct facet.
    for idx, _layout in enumerate(fig.layout.annotations):
        _attribute_name = _layout.text.replace("attribute_name=", "")
        _subset_by_attribute = agent_steps_table.loc[agent_steps_table["attribute_name"] == _attribute_name]
        fig.add_trace(
            go.Scatter(
                name="CI Upper",
                x=_subset_by_attribute[plot_kwargs["x"]],
                y=_subset_by_attribute["attribute_value_ci_upper"],
                line={"width": 0},
                mode="lines",
                fillcolor="rgba(68, 68, 68, 0.3)",
                showlegend=False,
                marker={"color": "#444"},
            ),
            col=1,
            row=idx + 1,
        )
        # Note: we "fill to next y". This will not work properly
        # when we plot more than one trace on a single figure. As such,
        # we limit the view of aggregated traces with the mean and CI
        # to one attribute per plot.
        fig.add_trace(
            go.Scatter(
                name="CI Lower",
                x=_subset_by_attribute[plot_kwargs["x"]],
                y=_subset_by_attribute["attribute_value_ci_lower"],
                line={"width": 0},
                mode="lines",
                fillcolor="rgba(68, 68, 68, 0.3)",
                fill="tonexty",
                showlegend=False,
                marker={"color": "#444"},
            ),
            col=1,
            row=idx + 1,
        )
    return fig


def update_figure_with_y_axes_view_options(fig: go.Figure, custom_y_axes_dict: dict) -> go.Figure:
    """Updates the figure with the y-axes selected and stored in the `sess_state.custom_y_axes_dict`

    Parameters
    ----------
    fig : go.Figure
        Existing plotly figure
    custom_y_axes_dict : Dict
        A dictionary where the key is the attribute name and the value is the y-axis prop

    Returns
    -------
    go.Figure
        Updated plotly figure
    """
    if len(custom_y_axes_dict) > 1:
        _yaxis_prop = {}
        for idx, _layout in enumerate(fig.layout.annotations):
            y_axis_name = "yaxis" if idx == 0 else f"yaxis{idx+1}"
            # Grab the y-axis properties according to the attribute name
            _prop = custom_y_axes_dict.get(_layout.text.replace("attribute_name=", ""))
            if _prop:
                _yaxis_prop[y_axis_name] = _prop

        fig.update_layout(**_yaxis_prop)
    else:
        _yaxis_prop = {"yaxis": custom_y_axes_dict[""]}
        fig.update_layout(**_yaxis_prop)

    return fig


def _expander_checkboxes_xaxes_options_and_filter(agent_steps_table: pd.DataFrame) -> tuple[pd.DataFrame, dict]:
    """Section for x-axis options (1) Display Relavtive to last step
    and (2) Filter on Steps.

    Parameters
    ----------
    agent_steps_table : pd.DataFrame
        The agent steps table from EpisodeArtifactTable or a subset of it.
    pd.DataFrame
        The potentially modified agent_steps_table
    Dict
        Initial plotly kwargs
    """

    plot_kwargs: dict = {}
    # #########################################################################
    # User selections for how to display plots: Use Expander to match Y-axis
    # options
    # #########################################################################
    hash_id = "_".join(agent_steps_table["agent_id"].unique().tolist())
    with st.expander("**X-Axis View Options**", expanded=True):
        # X-axis: Time or Steps?
        # time = steps / frame_rate, where frame_rate is an optional value
        # when the frame_rate is None, the time columns will be null
        if agent_steps_table["time"].isnull().sum() == 0:
            step_time_option = (
                st.radio(label="Axis Index", options=["step", "time"], horizontal=True, key=f"{hash_id}_xaxis_index") or "step"
            )

        columns = st.columns(4)
        # Changes the x-axis to be relative to last step / second. That
        # is, uses the `step_rel_to_end` index rather than `step.
        with columns[0]:
            if st.checkbox("Display relative to last step / second", key=f"{hash_id}_rel_to_end"):
                plot_kwargs.update({"x": f"{step_time_option}_rel_to_end"})
            else:
                plot_kwargs.update({"x": step_time_option})
        # Filters by step: allows the user to input a min / max
        # for the step range.
        with columns[1]:
            if st.checkbox("Filter on x-axis index", key=f"{hash_id}_filter_on_steps"):
                with columns[2]:
                    min_value = agent_steps_table[plot_kwargs["x"]].min()
                    max_value = agent_steps_table[plot_kwargs["x"]].max()

                    selected_min = st.number_input("start", min_value=min_value, max_value=max_value, key=f"{hash_id}_start")
                with columns[3]:
                    selected_max = st.number_input("end", min_value=min_value, max_value=max_value, value=max_value, key=f"{hash_id}_end")
                steps_filter = agent_steps_table[plot_kwargs["x"]].between(selected_min, selected_max, inclusive="both")
                agent_steps_table = agent_steps_table.loc[steps_filter]

    return agent_steps_table, plot_kwargs


def _expander_filter_dataframe(tables_dict: dict, filtered_agent_table: pd.DataFrame, sess_state: SessionStateProxy) -> pd.DataFrame:
    """
    Dataframe filter GUI
    """
    table_names = [
        name for name, pd_table in tables_dict.items() if "test_case" in pd_table.columns and name != FlattenedAgentStep.__name__
    ]
    _agent_id_key = filtered_agent_table["agent_id"].unique()[0]
    with st.expander("**Advanced: Filter Test Cases (Show All / Aggregate Only)**"):
        st.markdown(
            """ **Overview**: This section allows the user to filter down which test cases to show
                  when all the test cases are being displayed.
                  First choose which tables to filter on and press `DONE`.
                  The filters will stack together as `AND` operations.
                  A view of the filtered data for each table is shown for reference.
                  Press `Clear Filters` to clear all filters. """
        )
        if st.button("Clear Filters", key=f"clear_filter_{_agent_id_key}"):
            sess_state.tables_selected_for_filter = False
        with st.form(f"table_selections_{_agent_id_key}"):
            selected_tables = st.multiselect(
                label="Choose Which Tables to Filter on", options=table_names, key=f"multi_select_{_agent_id_key}"
            )
            tables_selected_button = st.form_submit_button("Done")
        if tables_selected_button or sess_state.get("tables_selected_for_filter"):
            # Start with all test cases in agent table
            filtered_test_cases = set(filtered_agent_table["test_case"].unique())
            for table_name in selected_tables:
                _dataframe = tables_dict[table_name].copy()
                _dataframe = ui_dataframe_filter(_dataframe, markdown_title=f"### Table: {table_name}")
                filtered_test_cases = filtered_test_cases.intersection(set(_dataframe["test_case"].unique()))
                sess_state.tables_selected_for_filter = True

            filtered_test_cases_as_list = list(filtered_test_cases)
            filtered_test_cases_as_list.sort()
            st.markdown(f"**Filtered Test Cases**: `{filtered_test_cases_as_list}`")

            idx = filtered_agent_table["test_case"].isin(filtered_test_cases)
            return filtered_agent_table.loc[idx]

    return filtered_agent_table


def _expander_set_yaxes_view_options(
    agent_steps_table: pd.DataFrame,
    agent_id: str,
    plot_layout: str,
    options_dict: dict[str, str],
    sess_state: SessionStateProxy,
    attributes_in_dataframe: list[str],
) -> SessionStateProxy:
    """Expander section to set y-axis bounds for each plot.

    Parameters
    ----------
    agent_steps_table : pd.DataFrame
        The agent steps table from EpisodeArtifactTable or a subset of it.
    plot_layout : str
        The radio selection output, this is a value of `options_dict`
    options_dict : Dict[str, str]
        The dictionary of options for how to display the plots
    sess_state : SessionStateProxy
        Streamlit session state
    attributes_in_dataframe : List[str]
        The selected attributes that are present in the dataframe, ordered.
        The order of this list will dictate the order of the y-axis attributes
        shown in this section.

    Returns
    -------
    SessionStateProxy
        Streamlit session state.
    """

    hash_id = "_".join(agent_steps_table["agent_id"].unique().tolist())
    with st.expander(label="**Y-Axis View Options**", expanded=False):
        # Option to show magnitude of y-axis values only.
        apply_abs_value = st.checkbox(label="Show Absolute Values", key=f"mag_{hash_id}")
        if apply_abs_value:
            agent_steps_table["attribute_value"] = agent_steps_table["attribute_value"].abs()
        st.markdown("Defaults to the min & max values (with a 5% buffer) across test-cases for each attribute.")
        with st.form(f"yaxis_options_{hash_id}"):
            # Only one y-axis when same plot is selected.
            iterate_over = [""] if plot_layout == options_dict["same_plot"] else attributes_in_dataframe
            _output_dict = {}
            default_min_y = agent_steps_table["attribute_value"].min() * 0.95
            default_max_y = agent_steps_table["attribute_value"].max() * 1.05
            for _attribute in iterate_over:
                st.markdown(f"{_attribute}")
                # When plotted on separate plots, min / max defaults are specific to each selected attributed
                if _attribute in attributes_in_dataframe:
                    default_min_y = agent_steps_table.loc[agent_steps_table["attribute_name"] == _attribute]["attribute_value"].min() * 0.95
                    default_max_y = agent_steps_table.loc[agent_steps_table["attribute_name"] == _attribute]["attribute_value"].max() * 1.05
                columns = st.columns(2)
                # NOTE: -0.05 / 1.05 when min / max values evaluate to 0 to create a 5% buffer for view
                with columns[0]:
                    min_y = st.number_input(
                        label="y-axis min", key=f"{hash_id}_{_attribute}_min", value=default_min_y if default_min_y != 0.0 else -0.05
                    )
                with columns[1]:
                    max_y = st.number_input(
                        label="y-axis max", key=f"{hash_id}_{_attribute}_max", value=default_max_y if default_max_y != 0.0 else 1.05
                    )
                _output_dict[_attribute] = {"range": [min_y, max_y]}
            if st.form_submit_button("Submit"):
                sess_state.custom_y_axes_dict[agent_id] = _output_dict.copy()

    return sess_state


def _plot_layout_aggregate_option(agent_steps_table: pd.DataFrame, plot_kwargs: dict) -> tuple[pd.DataFrame, dict]:
    """Appends the 95% CI columns to `agent_steps_table` and update `plot_kwargs` accordingly.

    Parameters
    ----------
    agent_steps_table : pd.DataFrame
        The agent steps table from EpisodeArtifactTable or a subset of it.
    plot_kwargs : dict
        The plotly kwargs thus far.

    Returns
    -------
    pd.DataFrame
        The modified `agent_steps_table`
    Dict
        The updated plot_kwargs
    """

    agent_steps_table = append_confidence_interval_columns(
        dataframe=agent_steps_table, steps_col=plot_kwargs["x"], attribute_name_col="attribute_name", attribute_value_col="attribute_value"
    )
    plot_kwargs.update({"facet_col": "attribute_name", "facet_col_wrap": 1, "y": "attribute_value_mean"})

    return agent_steps_table, plot_kwargs


def _plot_layout_different_plots_option(agent_steps_table: pd.DataFrame, plot_kwargs: dict) -> dict:
    """Updates the plot_kwargs

    Parameters
    ----------
    agent_steps_table : pd.DataFrame
        The agent steps table from EpisodeArtifactTable or a subset of it.
    plot_kwargs : dict
        The plotly kwargs thus far.

    Returns
    -------
    Dict
        The updated plot_kwargs
    """
    n_selected_attributes = agent_steps_table["attribute_name"].nunique()
    plot_kwargs.update(
        {
            "facet_col": "attribute_name",
            "facet_col_wrap": 1,
            "y": "attribute_value",
            "height": 300 * n_selected_attributes,
            "facet_row_spacing": 0.03,
        }
    )
    return plot_kwargs


def _plot_layout_same_plot_option(plot_kwargs: dict) -> dict:
    """Updates the plot_kwargs

    Parameters
    ----------
    plot_kwargs : dict
        The plotly kwargs thus far.

    Returns
    -------
    Dict
        The updated plot_kwargs
    """
    plot_kwargs.update({"color": "attribute_name", "y": "attribute_value", "height": 500})
    return plot_kwargs


def _plot_layout_show_all_option(agent_steps_table: pd.DataFrame, plot_kwargs: dict) -> dict:
    """Updates the plot_kwargs

    Parameters
    ----------
    agent_steps_table : pd.DataFrame
        The agent steps table from EpisodeArtifactTable or a subset of it.
    plot_kwargs : dict
        The plotly kwargs thus far.

    Returns
    -------
    Dict
        The updated plot_kwargs
    """
    n_selected_attributes = agent_steps_table["attribute_name"].nunique()
    plot_kwargs.update(
        {
            "color": "test_case",
            "y": "attribute_value",
            "facet_col": "attribute_name",
            "facet_col_wrap": 1,
            "height": 500 * n_selected_attributes,
        }
    )
    return plot_kwargs


def _section_headers():
    st.markdown("## Agent Steps")
    st.markdown(
        """
    - **Array elements**: Any observations or actions that were originally array elements are flattened in the table.
      For example, if an attribute named `obs_sensor` is originally: [0, 1, 0], it will be flattened to
      `obs_sensor_0 = 0`, `obs_sensor_1=1`,`obs_sensor_2=0` in the flattened table.
    - **Repeated Observations**: Repeated observations are stored in CoRL as a `List[Dict[str, Any]]`. Again, this is flattened
      in the table. For example, given a repeated observation sensor in the format:
      `ObserveSensorRepeated_SensorExample = [{"position": [1,2], "angle":90}, {"position": [10,20], "angle":45}]`,
      the corresponding on rows in the table are:
        ```
        # These all correspond to the first element of the top-level list
        ObserveSensorRepeated_SensorExample_position[0]_0 = 1 # First element of position
        ObserveSensorRepeated_SensorExample_position[0]_1 = 2 # Second element of position
        ObserveSensorRepeated_SensorExample_angle[0] = 90

        # Second element of the top-level list
        ObserveSensorRepeated_SensorExample_position[1]_0 = 10 # First element of position
        ObserveSensorRepeated_SensorExample_position[1]_1 = 20 # Second element of position
        ObserveSensorRepeated_SensorExample_angle[0] = 45
        ```
    - **Multiplatform Agents**: The sensors and controls for multiplatform agents will have the platform name in the part name.
    """
    )



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Controller Glue
"""
import logging
from collections import OrderedDict
from functools import cached_property

from corl.glues.base_glue import BaseAgentControllerGlue, BaseAgentPlatformGlueValidator
from corl.libraries.env_space_util import EnvSpaceUtil
from corl.libraries.property import DictProp
from corl.libraries.units import Quantity, corl_get_ureg
from corl.simulators.base_parts import BaseController
from corl.simulators.common_platform_utils import get_controller_by_name

code_logger = logging.getLogger("code_timing")


class ControllerGlueValidator(BaseAgentPlatformGlueValidator):
    """
    controller: Which controller to get from parent platform
    """

    controller: str


class ControllerGlue(BaseAgentControllerGlue):
    """
    This simple glue class wraps a controller and creates an action space based on the controller
    This class has no observation space or observations
    """

    def __init__(self, **kwargs) -> None:
        self.config: ControllerGlueValidator
        super().__init__(**kwargs)

        self._controller = get_controller_by_name(self._platform, self.config.controller)
        self._controller_name: str = self.config.controller
        self._key = self._controller.control_properties.name
        self._control_properties = self._controller.control_properties

        control_properties = self._control_properties
        if isinstance(control_properties, list):
            controller_names: list[str] = [controller.name for controller in control_properties]
            unique_name = "".join([controller_name.capitalize() for controller_name in controller_names])
        else:
            assert control_properties.name is not None
            unique_name = control_properties.name.capitalize()

        self._uname = f"{self.config.controller}_{unique_name}"

    @property
    def controller(self) -> BaseController:
        """Returns controller

        Returns
        -------
        BaseController
            The controller for this glue
        """
        return self._controller

    @staticmethod
    def get_validator() -> type[ControllerGlueValidator]:
        return ControllerGlueValidator

    @cached_property
    def action_prop(self):
        return self._control_properties

    @cached_property
    def observation_prop(self):
        tmp = {"control": self._control_properties}
        return DictProp(spaces=tmp)

    def apply_action(
        self, action: EnvSpaceUtil.sample_type, observation: EnvSpaceUtil.sample_type, action_space, obs_space, obs_units
    ) -> None:
        """Apply the action for the controllers

        Parameters
        ----------
        action : EnvSpaceUtil.sample_type
            The action that is to be applied at the controller level
        observation : EnvSpaceUtil.sample_type
            The current observable state by the agent (integration focus)
        """
        assert isinstance(action, Quantity) or (
            isinstance(action, dict) and all(isinstance(action_value, Quantity) for action_value in action.values())
        )
        self._controller.apply_control(control=action)

    def get_applied_control(self):
        if self._agent_removed:
            assert self.action_space is not None
            return corl_get_ureg().Quantity(self.action_space.sample(), self.action_prop.get_units())
        return self._controller.get_validated_applied_control()

    def get_observation(self, other_obs: OrderedDict, obs_space, obs_units) -> OrderedDict:
        obs_dict = OrderedDict()
        obs_dict["control"] = self.get_applied_control()
        return obs_dict

    def get_unique_name(self):
        """Provides a unique name of the glue to differentiate it from other glues."""
        return self._uname

    @property
    def resolved_controller_class_name(self) -> str:
        """Class name of the internal controller."""
        return type(self._controller).__name__


class ControllerGlueLegacy(ControllerGlue):
    """
    Version of ControllerGlue that matches the corl2 action space, where
    action space was a Dict
    """

    @cached_property
    def action_prop(self):
        tmp = {self._key: super().action_prop}
        return DictProp(spaces=tmp)

    def apply_action(self, action, observation, action_space, obs_space, obs_units):
        control = action[self._key]
        return super().apply_action(control, observation, action_space, obs_space, obs_units)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Entry / Launch point to run the storage process.
"""

import typing

import jsonargparse

from corl.evaluation.util.storage import Storage


def get_args(path: str | None = None) -> jsonargparse.Namespace:
    """
    Obtain running arguments for the storage process.

    Parameters
    ----------
    path : str
        optional path to a yaml config containing running arguments

    Returns
    -------
    instantiated: jsonargparse.Namespace
        namespace object that contains parsed running arguments
    """

    parser = jsonargparse.ArgumentParser()
    parser.add_argument("--cfg", action=jsonargparse.ActionConfigFile, help="the means by which to specify a yaml config to storage")
    parser.add_argument("--storage_utility", type=list[Storage], help="storage class to use")
    parser.add_argument(
        "--artifacts_location_config",
        type=dict[str, typing.Any],
        help="where are the eval_data_location, metric_file_location, event_table_location, agent_checkpoints ",
    )
    args = parser.parse_path(path) if path else parser.parse_args()
    return parser.instantiate_classes(args)


def main(instantiated_args):
    """
    Core logic for the storage process.

    Parameters
    ----------
    instantiated_args: jsonargparse.Namespace
        contains running arguments for the storage process.
    """
    for storage in instantiated_args["storage_utility"]:
        storage.load_artifacts_location(instantiated_args["artifacts_location_config"])
        storage.store()


def pre_main():
    """
    calls gets current args and passes them to main
    """
    instantiated, _ = get_args()
    main(instantiated)


if __name__ == "__main__":
    pre_main()



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

from abc import abstractmethod


class BasePlatformType:
    """
    Class describing a type of platform in a simulation
    This object will be used to register parts to the plugin library
    and help the simulator properly initialize platforms that may have
    custom methods of initialization
    """

    @classmethod
    @abstractmethod
    def match_model(cls, config: dict) -> bool:
        """
        this function takes in the platform configuration
        description for this platform and must return True
        if the config dict should be of this platform type
        or false otherwise
        """



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Sensors for GymnasiumSimulator
"""
import typing
from collections import OrderedDict
from functools import cached_property

import numpy as np

from corl.glues.base_dict_wrapper import BaseDictWrapperGlue
from corl.libraries.property import BoxProp, DictProp


class ProjectedQuantity(BaseDictWrapperGlue):
    """
    Glue that takes a quantity and a number of angles and projects the quantity in the direction of the angles given.
    """

    class Fields:
        """
        Fields in this glue
        """

        PROJECTED_QUANTITY = "projected_quantity"

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        if "quantity" not in self.glues():
            raise KeyError("Missing key: quantity")

        wrapped_name = self.glues()["quantity"].get_unique_name()

        self._uname: str = ""
        if wrapped_name is None:
            self._uname = "ProjectedQuantity"
        else:
            self._uname = f"Projected_{wrapped_name}"

    def get_unique_name(self) -> str:
        return self._uname

    @cached_property
    def observation_prop(self):
        if "quantity" not in self.glues():
            raise KeyError("Missing key: quantity")

        tmp = self.glues()["quantity"].observation_prop
        if not isinstance(tmp, DictProp):
            raise RuntimeError("projected_quantity glue received an invalid Prop from the glue it is wrapping")
        wrapped_space = next(iter(tmp.spaces.values()))
        max_wrapped_obs = np.amax(np.array([np.abs(wrapped_space.low), np.abs(wrapped_space.high)]))

        return DictProp(
            spaces={self.Fields.PROJECTED_QUANTITY: BoxProp(low=[-max_wrapped_obs], high=[max_wrapped_obs], unit=wrapped_space.get_units())}
        )

    def get_observation(self, other_obs: OrderedDict, obs_space, obs_units) -> typing.OrderedDict[str, np.ndarray]:
        d = OrderedDict()

        observations = {
            k: next(iter(v.get_observation(other_obs, obs_space, obs_units).values())).m for k, v in self.glues().items()  # type: ignore
        }

        projected_value = observations.pop("quantity")
        projected_value *= np.prod(np.cos(list(observations.values())))

        d[self.Fields.PROJECTED_QUANTITY] = self.observation_prop[self.Fields.PROJECTED_QUANTITY].create_quantity(projected_value)
        return d



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import difflib
import gc
import pprint
from functools import partial
from pathlib import Path
from typing import Annotated
from weakref import CallableProxyType, proxy

from pydantic import BaseModel, BeforeValidator, ConfigDict, Field
from ray.rllib.algorithms.algorithm import Algorithm

from corl.evaluation.evaluation_factory import FactoryLoader
from corl.evaluation.recording.i_recorder import IRecorder
from corl.evaluation.runners.section_factories.engine.rllib.rllib_trainer import RllibConfig
from corl.evaluation.runners.section_factories.plugins.plugins import Plugins
from corl.evaluation.runners.section_factories.task import Experiment, Task
from corl.evaluation.runners.section_factories.teams import LoadableCorlAgent, Teams
from corl.evaluation.runners.section_factories.test_cases.test_case_manager import NoTestCases, TestCaseStrategy
from corl.libraries.algorithm_helper import cleanup_algorithm
from corl.libraries.factory import Factory


def factory(v, info, **kwargs):
    """Factory helper function to auto-generate objects"""
    obj_type = kwargs["obj_type"]
    if isinstance(v, obj_type):
        return v

    try:
        return Factory.resolve_factory({"type": obj_type, "config": v}, info=info)
    except ValueError:
        raise
    except Exception as e:  # noqa: BLE001
        raise ValueError(f"{e}") from e


class EvalExperiment(BaseModel):
    """Configures/validates inference setup"""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    teams: Annotated[Teams, BeforeValidator(partial(factory, obj_type=Teams))]
    task: Annotated[Task, BeforeValidator(partial(factory, obj_type=Task))]
    plugins: Annotated[Plugins, BeforeValidator(partial(factory, obj_type=Plugins))] | None = None
    engine: Annotated[RllibConfig, BeforeValidator(partial(factory, obj_type=RllibConfig))]
    test_case_manager: Annotated[TestCaseStrategy, BeforeValidator(Factory.resolve_factory)] = Field(default_factory=NoTestCases)

    recorders: list[Annotated[IRecorder, BeforeValidator(FactoryLoader.resolve_factory)]] = Field(default=[], validate_default=True)

    cfg: Path | None = None
    output_dir: Path | None = None

    experiment_: Experiment | None = None
    algorithm_: Algorithm | None = None

    @property
    def experiment(self) -> Experiment:
        if not self.experiment_:
            self.experiment_ = Experiment(teams=self.teams, task=self.task)  # type: ignore
        return self.experiment_

    @property
    def algorithm(self) -> CallableProxyType:
        if not self.algorithm_:
            raise RuntimeError("Algorithm not created")
        return proxy(self.algorithm_)

    def __enter__(self):
        if self.algorithm_ is None:
            """Build an algorithm from engine/experiment/plugins"""

            try:
                algorithm = self.engine.rllib.generate(
                    self.experiment,
                    test_case_manager=self.test_case_manager,
                    plugins=self.plugins,
                )
            except ValueError:
                raise
            except Exception as e:  # noqa: BLE001
                raise ValueError from e

            # Now insert each participant's weights into the algorithm
            def apply_weights(agent: LoadableCorlAgent):
                try:
                    agent.agent_loader.apply_to_algorithm(algorithm, agent.name)
                except Exception as e:  # noqa: BLE001
                    err = f"{e}"
                    if agent.agent_loader.env_config:
                        trained_glues = agent.agent_loader.env_config["agents"][agent.agent_loader.agent_id].class_config.config["glues"]
                        inference_glues = algorithm.config["env_config"]["agents"][agent.name].class_config.config["glues"]

                        trained_glue_config = pprint.pformat(trained_glues).split("\n")
                        inference_glue_config = pprint.pformat(inference_glues).split("\n")

                        html_diff = difflib.HtmlDiff().make_file(trained_glue_config, inference_glue_config)
                        text_diff = "\n".join(difflib.ndiff(trained_glue_config, inference_glue_config))

                        diff_path = Path(f"{agent.name}_diff.html")
                        diff_path.write_text(html_diff, encoding="UTF-8")
                        err += f"\nAgent config changes between training and inference for {agent.name} saved to '{diff_path}'\n{text_diff}"

                    raise ValueError(f"{err}") from e

            self.teams.iterate_on_participant(apply_weights)
            if algorithm.workers is not None:
                algorithm.workers.sync_weights()

            self.algorithm_ = algorithm

        return self.algorithm

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.algorithm_ is not None:
            cleanup_algorithm(self.algorithm_)
            self.algorithm_ = None
            gc.collect()



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Module with implementation for multiple Observations
"""
import logging

from corl.rewards.base_measurement_operation import ExtractorSet, ObservationExtractorValidator
from corl.rewards.reward_func_base import RewardFuncBase, RewardFuncBaseValidator


class MultiMeasurementOperationValidator(RewardFuncBaseValidator):
    """
    observations: Dict of dicts of observation extractor arguments described in ObservationExtractorValidator
    """

    observations: dict[str, ObservationExtractorValidator]


class MultiMeasurementOperation(RewardFuncBase):
    """Base class for any reward that is to operate on multiple measurements of some kind"""

    @staticmethod
    def get_validator() -> type[MultiMeasurementOperationValidator]:
        return MultiMeasurementOperationValidator

    def __init__(self, **kwargs) -> None:
        self.config: MultiMeasurementOperationValidator
        super().__init__(**kwargs)
        self._logger = logging.getLogger(self.name)
        # construct extractors
        self.extractors: dict[str, ExtractorSet] = {}
        for key, observation in self.config.observations.items():
            self.extractors[key] = observation.construct_extractors()



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Reward that uses episode state to accumulate reward"""
import abc
import enum
import typing
from collections import OrderedDict
from functools import partial
from typing import Annotated

import gymnasium
import numpy as np
from pydantic import AfterValidator, BaseModel, NonNegativeFloat, PositiveFloat

from corl.dones.done_func_base import DoneStatusCodes
from corl.rewards.reward_func_base import RewardFuncBase, RewardFuncBaseValidator
from corl.simulators.base_simulator_state import BaseSimulatorState


class NegativeExponentialScaling(BaseModel):
    """Validation entry for negative exponential scaling."""

    scale: NonNegativeFloat
    eps: PositiveFloat


class EpisodeDoneRewardValidator(RewardFuncBaseValidator):
    """
    consolidate: if this done condition should attempt to reduce down to WLD in the case of
        multiple done conditions being set, with Win taking higher precedence
    skip_win_lose_sanity_check: notes if the done condition should skip the sanity check checking for
        both WIN and LOSS in an episodes done results, in case it was intentional
    """

    consolidate: bool = False
    skip_win_lose_sanity_check: bool = True


class EpisodeDoneReward(RewardFuncBase):
    """Base class for rewards that give rewards based on done information."""

    def __init__(self, **kwargs) -> None:
        self.config: EpisodeDoneRewardValidator
        super().__init__(**kwargs)
        self._counter = 0
        self._already_recorded: set[str] = set()
        self._status_codes: dict[DoneStatusCodes, list[str]] = {x: [] for x in DoneStatusCodes}

    @staticmethod
    def get_validator() -> type[EpisodeDoneRewardValidator]:
        return EpisodeDoneRewardValidator

    @staticmethod
    def exp_scaling(x, *, scale: float, eps: float) -> float:
        """Scale as a negative exponential."""
        return scale * np.exp(-np.abs(x / eps))

    @staticmethod
    def constant_scaling(_, *, scale: float) -> float:
        """Scale by a constant."""
        return scale

    @staticmethod
    def zero_scaling(_) -> float:
        """Scale as zero."""
        return 0

    @abc.abstractmethod
    def get_scaling_method(self, done_name, done_code) -> typing.Callable[[tuple[int, float]], float]:
        """Get the scaling method for a particular done name and code."""
        raise NotImplementedError()

    def __call__(
        self,
        observation: OrderedDict,
        action,
        next_observation: OrderedDict,
        state: BaseSimulatorState,
        next_state: BaseSimulatorState,
        observation_space: gymnasium.Space,
        observation_units: OrderedDict,
    ) -> float:
        reward: float = 0.0

        done_state = next_state.agent_episode_state.get(self.config.agent_name, {})
        for done_name, done_code in done_state.items():
            if done_name in self._already_recorded:
                continue
            self._already_recorded.add(done_name)
            self._status_codes[done_code].append(done_name)

        if (
            not self.config.skip_win_lose_sanity_check
            and len(self._status_codes[DoneStatusCodes.WIN]) > 0
            and len(self._status_codes[DoneStatusCodes.LOSE]) > 0
        ):
            raise RuntimeError(
                "EpisodeDoneReward found both WIN and LOSS set during this episode, if this is intended set skip_sanity_check=True"
            )

        # this will loop starting from win and go down
        consolidate_break = False
        for done_status in DoneStatusCodes:
            for done_name in self._status_codes[done_status]:
                reward += self.get_scaling_method(done_name, done_status)(self._counter)  # type: ignore[arg-type]
                if self.config.consolidate:
                    consolidate_break = True
                    break
            if consolidate_break:
                break

        self._counter += int(1 / next_state.sim_update_rate_hz)

        return reward


class EpisodeDoneStateRewardValidator(EpisodeDoneRewardValidator):
    """Validation for EpisodeDoneStateReward."""

    win: NegativeExponentialScaling | float = 0
    partial_win: NegativeExponentialScaling | float = 0
    draw: NegativeExponentialScaling | float = 0
    partial_loss: NegativeExponentialScaling | float = 0
    lose: NegativeExponentialScaling | float = 0


class EpisodeDoneStateReward(EpisodeDoneReward):
    """Reward that responds to done condition state, once per done condition triggered."""

    def __init__(self, **kwargs) -> None:
        self.config: EpisodeDoneStateRewardValidator
        super().__init__(**kwargs)

        self._status_code_func = {}
        for code in DoneStatusCodes:
            code_name = code.name.lower()
            if not hasattr(self.config, code_name):
                raise RuntimeError(f"Unknown done status code: {code_name}")
            code_data = getattr(self.config, code_name)
            if isinstance(code_data, NegativeExponentialScaling):
                self._status_code_func[code_name] = partial(self.exp_scaling, scale=code_data.scale, eps=code_data.eps)
            else:
                self._status_code_func[code_name] = partial(self.constant_scaling, scale=code_data)

    @staticmethod
    def get_validator() -> type[EpisodeDoneStateRewardValidator]:
        return EpisodeDoneStateRewardValidator

    def get_scaling_method(self, done_name, done_code) -> typing.Callable[[int], float]:  # type: ignore[override]
        return self._status_code_func[done_code.name.lower()]


class _MissingMethod(enum.Enum):
    zero = "zero"
    raise_exception = "raise"


def not_empty(v):
    """Ensure that some done condition is specified."""
    if len(v) == 0:
        raise ValueError("Rewarded dones cannot be empty")
    return v


class EpisodeDoneNameRewardValidator(EpisodeDoneRewardValidator):
    """Validation for EpisodeDoneNameReward."""

    rewarded_dones: Annotated[dict[str, NegativeExponentialScaling | float], AfterValidator(not_empty)] = {}
    missing_method: _MissingMethod = _MissingMethod.zero


class EpisodeDoneNameReward(EpisodeDoneReward):
    """Reward that responds once to individual dones."""

    def __init__(self, **kwargs) -> None:
        self.config: EpisodeDoneNameRewardValidator
        super().__init__(**kwargs)

        self._done_name_func = {}
        for name, args in self.config.rewarded_dones.items():
            if isinstance(args, NegativeExponentialScaling):
                self._done_name_func[name] = partial(self.exp_scaling, scale=args.scale, eps=args.eps)
            else:
                self._done_name_func[name] = partial(self.constant_scaling, scale=args)

    @staticmethod
    def get_validator() -> type[EpisodeDoneNameRewardValidator]:
        return EpisodeDoneNameRewardValidator

    def get_scaling_method(self, done_name, done_code) -> typing.Callable[[int], float]:  # type: ignore[override]
        if self.config.missing_method == _MissingMethod.zero:
            return self._done_name_func.get(done_name, self.zero_scaling)

        return self._done_name_func[done_name]



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from abc import ABC, abstractmethod


class PlatformDeserializerPlotlyAnimation(ABC):
    """Abstract class to define how to deserialize a platform for plotting"""

    @abstractmethod
    def get_position(self, serialized_data: dict) -> tuple | list:
        """Deserialized position data. This tells plotly where to plot
        a point for each platform on one step.

        Parameters
        ----------
        serialized_data : Dict
            Serialized platform data.

        Returns
        -------
        Union[Tuple, List]
            Position for each step, stored as a list or tuple.
            2-points are expected for 2-dimensional day and 3-points
            are expected for 3-dimensional data.
        """
        raise NotImplementedError

    @abstractmethod
    def get_dimensionality(self, serialized_data: dict) -> int:
        """Returns the dimensionality of the world - expected to be
        2D or 3D

        Parameters
        ----------
        serialized_data : Dict
            _description_

        Returns
        -------
        int
            _description_

        Raises
        ------
        NotImplementedError
            _description_
        """
        raise NotImplementedError

    @abstractmethod
    def get_position_units(self, serialized_data: dict) -> tuple | list:
        """Deserialized position units. This tells plotly what the units of
        the position are. The units are shown in the animation.

        Parameters
        ----------
        serialized_data : Dict
            Serialized platform data.

        Returns
        -------
        Union[Tuple,List]
            A tuple of units.
        """
        raise NotImplementedError

    @abstractmethod
    def get_platform_name(self, serialized_data: dict) -> str:
        """Returns the platform name

        Parameters
        ----------
        serialized_data : Dict
            Serialized platform data.

        Returns
        -------
        str
            Platform name
        """
        raise NotImplementedError

    @abstractmethod
    def get_metadata_for_text_display(self, serialized_data) -> dict[str, int | (float | str)]:
        """Deserializes metadata into a flattened dictionary. This is anything
        you want to include as optional text to displayed above the platform.

        Parameters
        ----------
        serialized_data : Dict
            Serialized platform data.

        Returns
        -------
        Dict[str, Union[int, float, str]]
            Any information you want to include in a flattened dictionary
            that could be displayed above the platform. This could be the velocity,
            heading, applied controls, sensors, etc.
        """
        raise NotImplementedError



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import pathlib
import pickle
from abc import abstractmethod
from pathlib import Path

from corl.evaluation.evaluation_artifacts import EvaluationArtifact_Metrics, EvaluationArtifact_Visualization
from corl.evaluation.scene_processors import SceneProcessors


class Visualization:
    """Base class for a visualization"""

    def __init__(self):
        pass

    def load(self, metrics_location: EvaluationArtifact_Metrics, visualization_location: EvaluationArtifact_Visualization):
        """
        Load a config detailing the location of the post_processed_data location and the save location for visualizations

        Parameters
        ----------
        config: dict
            dictionary detailing location of post_processed data and save location

        """
        if isinstance(metrics_location.location, str):
            post_processed_data_location = pathlib.Path(metrics_location.location).joinpath(metrics_location.file)
            with open(post_processed_data_location, "rb") as fp:
                self._post_processed_data: SceneProcessors = pickle.load(fp)  # noqa: S301
        else:
            raise RuntimeError("Pulling metrics from anything other than folder not supported")

        if isinstance(visualization_location.location, str):
            self.out_folder = Path(visualization_location.location)
        else:
            raise RuntimeError("Saving visualization to anything aside folder not supported")

    @abstractmethod
    def visualize(self):
        """Execute visualization code"""



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
property definitions
"""
from typing import Annotated

import numpy as np
from pydantic import Field, StrictFloat, StrictStr

from corl.libraries.property import BoxProp


class LatLonProp(BoxProp):
    """
    Lat Lon space
    """

    name: str = "LatLon"
    low: Annotated[list[StrictFloat], Field(min_length=2, max_length=2)] = [-90.0, -180.0]
    high: Annotated[list[StrictFloat], Field(min_length=2, max_length=2)] = [90.0, 180.0]
    unit: StrictStr = "degree"
    description: str = "Lat Lon"


class AltitudePropMeters(BoxProp):
    """
    Altitude space meters
    """

    name: str = "altitude"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [0.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [0.0]
    unit: StrictStr = "meter"
    description: str = "true altitude of the platform in meters. Altitude is measured above WGS-84 ellipsoid"


class AltitudeFeetProp(BoxProp):
    """
    Altitude space ft
    """

    name: str = "altitude"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [0.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [0.0]
    unit: StrictStr = "foot"
    description: str = "true altitude of the platform in ft"


class AltitudeMeterProp(BoxProp):
    """
    Altitude space meters ft
    """

    name: str = "altitude_m"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [0.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [0.0]
    unit: StrictStr = "meter"
    description: str = "true altitude of the platform in m"


class AltitudeRateProp(BoxProp):
    """
    Altitude rate space
    """

    name: str = "altitude_rate"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [0.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [0.0]
    unit: StrictStr
    description: str = "true altitude rate of the platform"


class OrientationProp(BoxProp):
    """
    Orientation space
    """

    name: str = "orientation"
    low: Annotated[list[StrictFloat], Field(min_length=3, max_length=3)] = [-np.pi] * 3
    high: Annotated[list[StrictFloat], Field(min_length=3, max_length=3)] = [np.pi] * 3
    unit: StrictStr = "rad"
    description: str = "yaw/heading, pitch, roll. Orientation is relative to the NED frame"


class OrientationRateProp(BoxProp):
    """
    Orientation rate space
    """

    name: str = "orientation_rate"
    low: Annotated[list[StrictFloat], Field(min_length=3, max_length=3)] = [-3 * np.pi] * 3
    high: Annotated[list[StrictFloat], Field(min_length=3, max_length=3)] = [3 * np.pi] * 3
    unit: StrictStr = "rad/s"
    description: str = "yaw rate, pitch rate, roll rate"


class FuelProp(BoxProp):
    """
    Fuel space in fuel / total fuel for the platform
    """

    name: str = "fuel_percentage"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [0.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [1.0]
    unit: StrictStr = "dimensionless"
    description: str = "ratio of remaining fuel / total fuel, 0 is empty and 1 is full"


class MachProp(BoxProp):
    """
    Mach speed space
    """

    name: str = "speed_mach"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [0.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [1.5]
    unit: StrictStr = "dimensionless"
    description: str = "speed of the platform in Mach"


class KcasProp(BoxProp):
    """
    Kcas speed space
    """

    name: str = "speed_kcas"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [0.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)]
    unit: StrictStr = "kcas"
    description: str = "KCAS (in knots)"


class KtasProp(BoxProp):
    """
    Ktas speed space
    """

    name: str = "speed_ktas"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [0.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)]
    unit: StrictStr = "knot"
    description: str = "KTAS (in knots)"


class KiasProp(BoxProp):
    """
    Kias speed space
    """

    name: str = "speed_kias"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [0.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)]
    unit: StrictStr = "kias"
    description: str = "KIAS (in knots)"


class TrueAirSpeedProp(BoxProp):
    """
    True airspeed space
    """

    name: str = "true_air_speed_fts"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [0.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)]
    unit: StrictStr = "foot / second"
    description: str = "true airspeed of the platform in foot per second"


class VelocityNEDProp(BoxProp):
    """
    Velocity NED space
    """

    name: str = "velocity_ned"
    low: Annotated[list[StrictFloat], Field(min_length=3, max_length=3)] = [-3000.0] * 3
    high: Annotated[list[StrictFloat], Field(min_length=3, max_length=3)] = [3000.0] * 3
    unit: StrictStr = "meter / second"
    description: str = "velocity in true airspeed (m/s) along north, east, down axis"


class AccelerationNEDProp(BoxProp):
    """
    Acceleration NED space
    """

    name: str = "acceleration_ned"
    low: Annotated[list[StrictFloat], Field(min_length=3, max_length=3)] = [-3000.0] * 3
    high: Annotated[list[StrictFloat], Field(min_length=3, max_length=3)] = [3000.0] * 3
    unit: StrictStr = "m/s^2"
    description: str = "acceleration in true airspeed (m/s^2) along north, east, down axis"


class FlightPathAngleProp(BoxProp):
    """
    Flight path angle space in degrees
    """

    name: str = "flight_path_angle_deg"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [-180.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [180.0]
    unit: StrictStr = "degree"
    description: str = "flight path angle of the platform in degrees"


class FlightPathAngleRadProp(BoxProp):
    """
    Flight path angle in radians
    """

    name: str = "flight_path_angle_rad"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [-np.pi / 2]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [np.pi / 2]
    unit: StrictStr = "rad"
    description: str = "flight path angle of the platform in rad"


class GLoadProp(BoxProp):
    """
    Gload space
    """

    name: str = "g_load"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [-20.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [20.0]
    unit: StrictStr = "standard_gravity"
    description: str = "G load on the platform"


class GloadNzProp(GLoadProp):
    """
    Gload space in z
    """

    name: str = "g_load NZ"


class GloadNyProp(GLoadProp):
    """
    Gload space in y
    """

    name: str = "g_load NY"


class GloadNxProp(GLoadProp):
    """
    Gload space in x
    """

    name: str = "g_load NX"


class AngleOfAttackProp(BoxProp):
    """
    Angle of attack space
    """

    name: str = "angle_of_attack"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [-180.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [180.0]
    unit: StrictStr = "degree"
    description: str = "The angle of attack of the platform in degrees"


class AngleOfAttackRateProp(BoxProp):
    """
    Angle of attack rate space
    """

    name: str = "angle_of_attack_rate"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [-180.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [180.0]
    unit: StrictStr = "deg/s"
    description: str = "The angle of attack rate of the platform in degrees/s"


class AngleOfSideSlipProp(BoxProp):
    """
    Angle of slide slip
    """

    name: str = "angle_of_side_slip"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [-90.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [90.0]
    unit: StrictStr = "degree"
    description: str = "The angle of side slip of the platform in degrees"


class AngleOfSideSlipRateProp(BoxProp):
    """
    Angle of slide slip rate space
    """

    name: str = "angle_of_side_slip_rate_dps"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [-180.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [180.0]
    unit: StrictStr = "deg/s"
    description: str = "The angle of side slip rate of the platform in degrees/sec"


class FuelWeightProp(BoxProp):
    """
    Fuel weight space
    """

    name: str = "fuel_weight_lbs"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [0.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [10000.0]
    unit: StrictStr = "lbs"
    description: str = "Fuel weight"


class WindDirectionProp(BoxProp):
    """
    Wind direction space
    """

    # https://docs.google.com/spreadsheets/d/1L7D4uqVQzY7rODqtnumB0Kv0-veQ1bHIItLWWrpOOmA/edit#gid=0
    name: str = "wind_direction_deg"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [0.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [360.0]
    unit: StrictStr = "degree"
    description: str = "Wind Direction in Degrees of the Platform"


class WindSpeedProp(BoxProp):
    """
    Wind speed space
    """

    # https://docs.google.com/spreadsheets/d/1L7D4uqVQzY7rODqtnumB0Kv0-veQ1bHIItLWWrpOOmA/edit#gid=0
    name: str = "wind_speed_kts"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [0.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [200.0]
    unit: StrictStr = "kts"
    description: str = "Wind speed in kts of the Platform"


class YawProp(BoxProp):
    """
    Yaw space
    """

    name: str = "orientation_yaw"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [-180.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [180.0]
    unit: StrictStr = "degree"
    description: str = "Get RPM Yaw (deg)"


class PitchProp(BoxProp):
    """
    Pitch space
    """

    name: str = "orientation_pitch"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [-180.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [180.0]
    unit: StrictStr = "degree"
    description: str = "Get RPM Pitch (deg)"


class RollProp(BoxProp):
    """
    Roll space
    """

    name: str = "orientation_roll"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [-180.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [180.0]
    unit: StrictStr = "degree"
    description: str = "Get RPM Roll (deg)"


class YawRateProp(BoxProp):
    """
    Yaw rate space
    """

    name: str = "orientation_yaw_rate"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [-3 * 180.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [3 * 180.0]
    unit: StrictStr = "deg/s"
    description: str = "Get RPM yaw rate (deg/sec)"


class PitchRateProp(BoxProp):
    """
    Pitch rate space
    """

    name: str = "orientation_pitch_rate"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [-3 * 180.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [3 * 180.0]
    unit: StrictStr = "deg/s"
    description: str = "Get RPM pitch rate (deg/sec)"


class RollRateProp(BoxProp):
    """
    Roll rate space
    """

    name: str = "orientation_roll_rate"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [-540.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [540.0]
    unit: StrictStr = "deg/s"
    description: str = "Get RPM roll rate (deg/sec)"


class HeadingProp(BoxProp):
    """
    default heading space
    """

    name: str = "heading"
    low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [-180.0]
    high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = [180.0]
    unit: StrictStr = "degree"
    description: str = "Direct (in degrees) the target heading"


def property_factor(prefix_name, input_property_class, low_prop=None, high_prop=None) -> type:
    """
    factory for creating space mods...
    """
    if issubclass(input_property_class, BoxProp):
        if low_prop is not None and high_prop is not None:

            class NewProperty(input_property_class):
                """
                dynamic class creation
                """

                low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = low_prop
                high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = high_prop

            return type(prefix_name + input_property_class.__name__, (NewProperty,), {})

        if low_prop is not None:

            class NewProperty(input_property_class):  # type: ignore[no-redef]
                """
                dynamic class creation
                """

                low: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = low_prop

            return type(prefix_name + input_property_class.__name__, (NewProperty,), {})

        if high_prop is not None:

            class NewProperty(input_property_class):  # type: ignore[no-redef]
                """
                dynamic class creation
                """

                high: Annotated[list[StrictFloat], Field(min_length=1, max_length=1)] = high_prop

            return type(prefix_name + input_property_class.__name__, (NewProperty,), {})

        class NewProperty(input_property_class):  # type: ignore[no-redef]
            """
            dynamic class creation
            """

        return type(prefix_name + input_property_class.__name__, (NewProperty,), {})

    raise NotImplementedError(f"Did not implement for target base type --- {input_property_class.__bases__}")



from abc import abstractmethod
from collections.abc import Iterator
from functools import partial
from pathlib import Path
from typing import Annotated, Any

from pydantic import BaseModel, BeforeValidator, ConfigDict, Field
from ray.rllib.policy.sample_batch import MultiAgentBatch, SampleBatch

from corl.evaluation.launchers.evaluate_validator import EvalExperiment
from corl.libraries.factory import Factory
from corl.parsers.yaml_loader import load_file


def get_default_evaluator() -> "BaseEvaluator":
    """Get the default evaluator type"""
    from corl.evaluation.runners.iterate_test_cases import TestCaseEvaluator

    return TestCaseEvaluator()


def get_default_processors() -> list["BaseProcessor"]:
    """Get the default processor type"""
    from corl.evaluation.runners.iterate_test_cases import ConfigSaver, SampleBatchProcessor

    return [SampleBatchProcessor(), ConfigSaver()]


def load_raw_config(v, info, **kwargs):
    if v is None or (isinstance(v, dict) and len(v) == 0):
        if info.data["path"] is None:
            raise ValueError("'path' or 'config' must be set.")
        return load_file(config_filename=info.data["path"])
    return v


def from_raw_config(key, v, info, **kwargs):
    config = info.data["raw_config"].get(key, v)

    return Factory.resolve_factory(config, info, **kwargs)


def processors_from_raw_config(v, info, **kwargs):
    processors = info.data["raw_config"].get("processors", v)

    return [Factory.resolve_factory(processor, info, **kwargs) for processor in processors]


class EvalConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    path: Path | None = None
    raw_config: Annotated[dict | None, BeforeValidator(load_raw_config)] = None

    evaluator: Annotated["BaseEvaluator", BeforeValidator(partial(from_raw_config, "evaluator"))] = Field(
        default_factory=get_default_evaluator, validate_default=True
    )
    processors: Annotated[list["BaseProcessor"], BeforeValidator(processors_from_raw_config)] = Field(
        default_factory=get_default_processors, validate_default=True
    )

    experiment: Annotated[EvalExperiment, BeforeValidator(partial(from_raw_config, "experiment"))] = Field(  # type: ignore
        default=None, validate_default=True
    )


class BaseEvaluatorValidator(BaseModel):
    """Base class validator for BaseEvaluator"""

    ...


class BaseEvaluator:
    """Base class for evaluator objects"""

    def __init__(self, **kwargs):
        self._config = self.get_validator()(**kwargs)

    @staticmethod
    def get_validator() -> type[BaseEvaluatorValidator]:
        """The validator for this class"""
        return BaseEvaluatorValidator

    @abstractmethod
    def __call__(self, config: EvalConfig) -> Iterator[SampleBatch | MultiAgentBatch]:
        """Generates results, which may be passed to a Processor"""
        ...

    @abstractmethod
    def reset(self):
        """Resets evaluation"""
        ...

    @abstractmethod
    def stop(self):
        """Stops evaluation"""
        ...


class BaseProcessorValidator(BaseModel):
    """Base class validator for BaseProcessor"""

    ...


class BaseProcessor:
    """Base class for result processor objects"""

    def __init__(self, **kwargs):
        self._config = self.get_validator()(**kwargs)

    @staticmethod
    def get_validator() -> type[BaseProcessorValidator]:
        """The validator for this class"""
        return BaseProcessorValidator

    @abstractmethod
    def __call__(self, config: EvalConfig, results: Any) -> Any:
        """Processes results from an Evaluator"""
        ...

    @abstractmethod
    def stop(self):
        """Stops the processor"""
        ...



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
from typing import Any

import numpy as np
import pandas as pd

from corl.evaluation.connection.base_epp_update import get_updates
from corl.evaluation.runners.inference.protocol.simple_epp_update import SimpleEppUpdate
from corl.evaluation.runners.section_factories.test_cases.base_network_strategy import BaseNetworkTestCaseStrategy
from corl.libraries.units import Quantity, Unit


class SimpleNetworkEppTestCaseStrategy(BaseNetworkTestCaseStrategy[SimpleEppUpdate]):
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

        self._cur_updates = SimpleEppUpdate(updates={})

    def _create_test_case_for_updates(self, updates: SimpleEppUpdate) -> pd.DataFrame | list[dict[str, Any]]:
        update_dict = get_updates(updates)

        test_case: dict[str, tuple[np.ndarray | float | int | str, Unit | str]] = {}
        value: np.ndarray | float | int | str
        units: Unit | str
        for key, qty in update_dict.items():
            if isinstance(qty, Quantity):
                value = qty.value
                units = qty.units
            else:
                value = qty
                units = "dimensionless"
            test_case[".".join(key)] = (value, units)

        return self._test_case + [test_case]  # fmt: skip  # noqa: RUF005



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Contains abstract interfaces defining how to record an Evaluation outcome
"""
import typing
from abc import abstractmethod
from pathlib import Path

from corl.evaluation.evaluation_outcome import EvaluationOutcome


class IRecord:
    """Interface that represents a record that can save and load EvaluationOutcomes"""

    @property
    @abstractmethod
    def absolute_path(self) -> Path:
        """Get the absolute path related to this record"""

    @abstractmethod
    def save(self, outcome: EvaluationOutcome) -> None:
        """Save evaluation outcome

        Arguments:
            outcome {EvaluationOutcome} -- Evaluation outcome to save
        """

    @abstractmethod
    def load(self) -> EvaluationOutcome:
        """Load EvaluationOutcome

        Returns:
            EvaluationOutcome -- Evaluation Outcome loaded from folder
        """


T = typing.TypeVar("T", bound=IRecord)


class IRecorder(typing.Generic[T]):
    """Interface the represents a recorder

    A recorder generates a Record
    """

    @abstractmethod
    def resolve(self) -> T:
        """Resolve the recorder and generate a Record"""



"""
This module defines the Docking1dSimulator class which maintains environment objects (platforms and entities)
and progresses a simulated training episode via the step() method.
"""


import numpy as np

from corl.libraries.plugin_library import PluginLibrary
from corl.simulators.base_simulator import BaseSimulator, BaseSimulatorResetValidator, BaseSimulatorState, BaseSimulatorValidator
from corl.simulators.docking_1d.entities import Deputy1D
from corl.simulators.docking_1d.platform import Docking1dPlatform


class Docking1dSimulatorValidator(BaseSimulatorValidator):
    """
    A validator for the Docking1dSimulatorValidator config.

    step_size: A float representing how many simulated seconds pass each time the simulator updates
    """

    step_size: float


class Docking1dSimulatorResetValidator(BaseSimulatorResetValidator):
    """
    A Validator for Docking1dSimulatorValidator reset configs

    Parameters
    ----------
    platform_config: dict
        Contains individual initialization dicts for each agent.
        Key is platform name, value is platform's initialization dict.
    """

    platform_config: dict | None = {}


class Docking1dSimulator(BaseSimulator):
    """
    Simulator for the 1D Docking task. Interfaces Docking1dPlatform with underlying Deputy1D entities in Docking
    simulation. Docking1dSimulator is responsible for initializing the platform objects for a simulation
    and knowing how to set up episodes based on input parameters from a parameter provider.
    It is also responsible for reporting the simulation state at each timestep.
    """

    @property
    def get_simulator_validator(self):
        return Docking1dSimulatorValidator

    @property
    def get_reset_validator(self):
        return Docking1dSimulatorResetValidator

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._state: BaseSimulatorState | None = None
        self.clock = 0.0

    def reset(self, config):
        config = self.get_reset_validator(**config)
        self.clock = 0.0

        # construct entities ("Gets the platform object associated with each simulation entity.")
        self.sim_entities = {}
        for agent_id in self.config.agent_configs:
            agent_reset_config = config.platforms.get(agent_id, {})
            self.sim_entities[agent_id] = Deputy1D(name=agent_id, **agent_reset_config)

        # construct platforms ("Gets the correct backend simulation entity for each agent.")
        sim_platforms = {}
        for agent_id, entity in self.sim_entities.items():
            agent_config = self.config.agent_configs[agent_id]
            sim_platforms[agent_id] = Docking1dPlatform(platform_name=agent_id, platform=entity, parts_list=agent_config.parts_list)

        self._state = BaseSimulatorState(sim_platforms=sim_platforms, sim_time=self.clock)
        self.update_sensor_measurements()
        return self._state

    def step(self, platforms_to_action):
        for agent_id, platform in self._state.sim_platforms.items():
            action = np.asarray(platform.get_applied_action().m, dtype=np.float32)
            entity = self.sim_entities[agent_id]
            entity.step(action=action, step_size=self.config.step_size)
            platform.sim_time = self.clock
        self.update_sensor_measurements()
        self.clock += self.config.step_size
        self._state.sim_time = self.clock
        return self._state

    @property
    def sim_time(self) -> float:
        return self.clock

    @property
    def platforms(self):
        return self._state.sim_platforms

    def update_sensor_measurements(self):
        """
        Update and cache all the measurements of all the sensors on each platform
        """
        for plat in self._state.sim_platforms.values():
            for sensor in plat.sensors.values():
                sensor.calculate_and_cache_measurement(state=self._state)

    def mark_episode_done(self, done_info, episode_state, metadata):
        pass

    def save_episode_information(self, dones, rewards, observations, observation_units):
        pass


# Register the Simulator with the PluginLibrary. Requires a class and a reference name.

PluginLibrary.AddClassToGroup(Docking1dSimulator, "Docking1dSimulator", {})



import typing
from os.path import expandvars
from pathlib import Path
from typing import Annotated

# from pydantic.validators import path_exists_validator
from pydantic import BeforeValidator, DirectoryPath, FilePath


def core_directory_path_validator(v: typing.Any):
    if isinstance(v, Path):
        v = str(v)
    if not isinstance(v, str):
        msg = "string or Path required as argument for Corl File/Directory Paths"
        raise TypeError(msg)
    vars_v = expandvars(v)
    return Path(Path.cwd(), vars_v).resolve() if vars_v.startswith(".") else Path(vars_v).resolve()


CorlFilePath = Annotated[FilePath, BeforeValidator(core_directory_path_validator)]
CorlDirectoryPath = Annotated[Path, BeforeValidator(core_directory_path_validator)]
CorlExistingDirectoryPath = Annotated[DirectoryPath, BeforeValidator(core_directory_path_validator)]



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Base Platform Abstract Class Object
"""
import abc

from pydantic import BaseModel, ConfigDict

from corl.simulators.base_parts import BaseController, BaseSensor


class BasePlatformValidator(BaseModel):
    """BasePlatformValidator

    Parameters
    ----------
        platform_name: str
            name of platform
        parts_list: typing.List[typing.Tuple]
            list of parts the agent uses to interact with the platform
    """

    platform_name: str
    parts_list: list[tuple]
    model_config = ConfigDict(arbitrary_types_allowed=True)


class BasePlatform(abc.ABC):
    """
    BasePlatform base abstraction for a platform object. Platforms can be aircraft, land vehicles,
    ground radar, satellites etc.
    Platforms have platform properties that describe the platform, for example position, velocity, etc.
    Platforms have platform parts which consist of controls or sensors.
    """

    def __init__(self, **kwargs) -> None:
        self.config: BasePlatformValidator = self.get_validator()(**kwargs)

        self._name = self.config.platform_name
        self._sensors = self._get_part_dict(self.config.parts_list, BaseSensor)
        self._controllers = self._get_part_dict(self.config.parts_list, BaseController)

        self.verify_unique_parts()

    @staticmethod
    def get_validator() -> type[BasePlatformValidator]:
        """
        get validator for this BasePlatform

        Returns:
            BasePlatformValidator -- validator the platform will use to generate a configuration
        """
        return BasePlatformValidator

    def _get_part_dict(self, part_class_list, part_base_class):
        """
        Get a list of the platform parts after being wrapped in the base integration classes.
        This will do the mapping from an SimSensor to a type of BaseSensor for example.
        Some parts have no matching sim counterpart and will not have None in the map instead
        of the part name

        Parameters
        ----------
        part_class_map: mapping of the part name to the pairs of base classes and config for that part
        parts: the parts we will wrap with the above mapping

        Returns
        -------
        A list of the wrapped platform parts
        """
        part_dict = {}
        for part_class, part_config in filter(lambda x: issubclass(x[0], part_base_class), part_class_list):
            tmp = part_class(self, part_config)
            if tmp.name in part_dict:
                if part_dict[tmp.name].config != tmp.config:
                    raise RuntimeError(
                        f"Platform {self._name} tried to add a part {tmp.name} to its part Dict for type {part_base_class}, "
                        f"however a part with this name already exists in the part dict with a different configuration.\n"
                        f"new part class was {part_class}, part config was {tmp.config},\n"
                        f"already existing part config is {part_dict[tmp.name].config}\n"
                        f"already existing part names are {part_dict.keys()}"
                    )
                continue
            part_dict[tmp.name] = tmp

        return part_dict

    def verify_unique_parts(self):
        """
        Verify all parts have a unique name
        """
        sensor_parts = set(self._sensors.keys())
        controller_parts = set(self._controllers.keys())
        part_union = sensor_parts.intersection(controller_parts)
        if filtered_intersection := {x for x in part_union if not isinstance(self._sensors[x], BaseController)}:
            raise RuntimeError(
                f"Parts on a platform need to have a unique name, but platform {self._name} "
                f"has the following sensors and controllers that share the same name: {filtered_intersection} "
                f"Parts may subclass both BaseSensor and BaseController and are not a concern in this check"
            )

    @property
    def name(self) -> str:
        """
        name the name of this object

        Returns
        -------
        str
            The name of this object
        """
        return self._name

    @property
    @abc.abstractmethod
    def operable(self) -> bool:
        """Is the platform operable?

        Returns
        -------
        bool
            Is the platform operable?
        """

    @property
    def sensors(self):
        """
        Sensors attached to this platform

        Returns
        ------
        List
            list of all sensors attached to this platform
        """
        return self._sensors

    @property
    def controllers(self):
        """
        Controllers attached to this platform

        Returns
        ------
        List
            list of all controllers attached to this platform
        """
        return self._controllers



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Base Simulator and Platform for Toy Openai Environments
This mainly shows a "how to use example" and provide an setup to unit test with
"""
import typing

from corl.libraries.plugin_library import PluginLibrary
from corl.libraries.units import Quantity
from corl.simulators.base_simulator import (
    AgentConfig,
    BaseSimulator,
    BaseSimulatorResetValidator,
    BaseSimulatorState,
    BaseSimulatorValidator,
)
from corl.simulators.pong.paddle_platform import PaddlePlatform, PaddleType
from corl.simulators.pong.pong import GameStatus, Pong


def strip_units(item):
    """Remove units form all values with units objects"""
    if isinstance(item, dict):
        for key, value in item.items():
            if isinstance(value, dict | Quantity):
                item[key] = strip_units(value)
        return item

    return item.m if isinstance(item, Quantity) else item


class PongSimulatorValidator(BaseSimulatorValidator):
    """
    A Validator for Docking1dSimulatorValidator reset configs

    Parameters
    ----------
    platform_config: dict
        Contains individual initialization dicts for each agent.
        Key is platform name, value is platform's initialization dict.
    enable_health: bool
        enable or disable health, which will turn collision of the paddle off and make it inoperable
        after touching the ball
    """

    platform_config: dict | None = {}
    paddle0_side: PaddleType = PaddleType.LEFT
    pong: Pong = Pong()

    agent_configs: typing.Mapping[str, AgentConfig] = None  # type: ignore


class PongSimulatorResetValidator(BaseSimulatorResetValidator):
    """
    A Validator for Docking1dSimulatorValidator reset configs

    Parameters
    ----------
    platform_config: dict
        Contains individual initialization dicts for each agent.
        Key is platform name, value is platform's initialization dict.
    enable_health: bool
        enable or disable health, which will turn collision of the paddle off and make it inoperable
        after touching the ball
    """

    platform_config: dict | None = {}
    paddle0_side: PaddleType = PaddleType.LEFT
    pong: Pong = Pong()

    agent_configs_reset: typing.Mapping[str, AgentConfig] = None  # type: ignore


class PongSimulatorState(BaseSimulatorState):
    """
    pong_game: the pong game being run
    game_status: the enum of the current game status
    """

    pong_game: Pong
    game_status: GameStatus = GameStatus.IN_PROGRESS


class PongSimulator(BaseSimulator):
    """
    Simulator backend for running pong game
    """

    def __init__(self, **kwargs) -> None:
        self.config: PongSimulatorValidator
        super().__init__(**kwargs)

        self._state: PongSimulatorState
        self._time = 0.0
        self.pong_game: Pong
        self._paddle0_type: PaddleType

    @property
    def get_simulator_validator(self) -> type[PongSimulatorValidator]:
        """
        access to afims validator
        """
        return PongSimulatorValidator

    @property
    def get_reset_validator(self) -> type[PongSimulatorResetValidator]:
        """Return validator"""
        return PongSimulatorResetValidator

    def reset(self, config: dict[str, typing.Any]):
        validated_config = self.get_reset_validator(**strip_units(config))
        if not validated_config.agent_configs_reset:
            validated_config.agent_configs_reset = self.config.agent_configs
        self.config.agent_configs = validated_config.agent_configs_reset
        self.config.paddle0_side = validated_config.paddle0_side

        self._time = 0.0

        self.pong_game = self.config.pong = validated_config.pong

        self._paddle0_type = PaddleType(self.config.paddle0_side)

        paddle_dict = self._initialize_paddles(validated_config.platforms)

        sim_platforms = {}
        for agent_id, agent_config in self.config.agent_configs.items():
            agent_config = self.config.agent_configs[agent_id]  # noqa: PLW2901
            sim_platforms[agent_id] = PaddlePlatform(
                platform_name=agent_id,
                parts_list=agent_config.parts_list,
                paddle_type=self._paddle_type(agent_id),
                paddle=paddle_dict[agent_id],
            )

        self._state = PongSimulatorState(sim_platforms=sim_platforms, sim_time=self._time, pong_game=self.pong_game)
        self.update_sensor_measurements()
        return self._state

    def step(self, platforms_to_action):
        keys = [platform.last_move for platform in self.platforms.values()]
        self._state.game_status = self.pong_game.step(keys)
        self._time += 1
        self._state.pong_game = self.pong_game
        self._state.sim_time = self._time
        self.update_sensor_measurements()
        return self._state

    def update_sensor_measurements(self):
        """
        Update and cache all the measurements of all the sensors on each platform
        """
        for plat in self._state.sim_platforms.values():
            for sensor in plat.sensors.values():
                sensor.calculate_and_cache_measurement(state=self._state)

    @property
    def sim_time(self) -> float:
        return self._time

    @property
    def platforms(self):
        return self._state.sim_platforms

    def mark_episode_done(self, done_info, episode_state, metadata):
        pass

    def save_episode_information(self, dones, rewards, observations, observation_units):
        pass

    def render(self, state, mode="human"):  # noqa: PLR6301
        """only render first environment"""
        return

    def _paddle_type(self, platform_id):
        if platform_id == "paddle0":
            return self._paddle0_type

        if platform_id == "paddle1":
            if self._paddle0_type == PaddleType.LEFT:
                return PaddleType.RIGHT
            if self._paddle0_type == PaddleType.RIGHT:
                return PaddleType.LEFT

        raise RuntimeError(f"Invalid paddle types for {platform_id}")

    def _get_agent_platform(self, platform_id):
        return self._pong_paddle(strip_units(self._paddle_type(platform_id)))

    def _initialize_paddles(self, platform_configs):
        if self._paddle0_type is PaddleType.LEFT:
            paddle0 = self.pong_game.left_paddle
            paddle1 = self.pong_game.right_paddle
        else:
            paddle0 = self.pong_game.right_paddle
            paddle1 = self.pong_game.left_paddle

        for platform, platform_config in platform_configs.items():
            if platform == "paddle0":
                paddle0 = paddle0.copy(update=strip_units(platform_config))
            if platform == "paddle1":
                paddle1 = paddle1.copy(update=strip_units(platform_config))

        if self._paddle0_type is PaddleType.LEFT:
            self.pong_game.left_paddle = paddle0
            self.pong_game.right_paddle = paddle1
        else:
            self.pong_game.left_paddle = paddle1
            self.pong_game.right_paddle = paddle0

        return {"paddle0": paddle0, "paddle1": paddle1}


PluginLibrary.AddClassToGroup(PongSimulator, "PongSimulator", {})



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Provides controllers to pass actions to wrapped  Gymnasium Environments
"""

import gymnasium
import numpy as np

from corl.libraries.plugin_library import PluginLibrary
from corl.libraries.property import BoxProp, DiscreteProp
from corl.libraries.units import Quantity, corl_get_ureg
from corl.simulators.base_parts import BaseController
from corl.simulators.gymnasium.gymnasium_available_platforms import GymnasiumAvailablePlatformTypeMain
from corl.simulators.gymnasium.gymnasium_simulator import GymnasiumSimulator


class GymnasiumMainController(BaseController):
    """
    GymnasiumController implementation for passing actions to wrapped  Gymnasium Environments
    """

    def __init__(self, parent_platform, config=None) -> None:
        act_space = parent_platform.action_space
        cont_prop: type[DiscreteProp] | type[BoxProp]

        if isinstance(act_space, gymnasium.spaces.Discrete):

            class DiscreteGymnasiumProp(DiscreteProp):
                """
                DiscreteGymnasiumProp can be updated via config and validate by pydantic
                """

                name: str = "gymnasium controller"
                unit: str = "dimensionless"
                n: int = act_space.n
                description: str = "gymnasium env action space"

            cont_prop = DiscreteGymnasiumProp

        elif isinstance(act_space, gymnasium.spaces.Box):

            class BoxGymnasiumProp(BoxProp):
                """
                BoxGymnasiumProp can be updated via config and validate by pydantic
                """

                name: str = "gymnasium controller"
                low: list[float] = act_space.low.tolist()
                high: list[float] = act_space.high.tolist()
                dtype: np.dtype = act_space.dtype
                unit: str = "dimensionless"
                description: str = "gymnasium env action space"

            cont_prop = BoxGymnasiumProp
        else:
            raise RuntimeError(f"This controller does not currently know how to handle a {type(act_space)} action space")

        super().__init__(parent_platform=parent_platform, config=config, property_class=cont_prop)
        new_action = corl_get_ureg().Quantity(self.control_properties.create_space().sample(), "dimensionless")
        parent_platform.save_action_to_platform(new_action)

    def apply_control(self, control) -> None:
        assert isinstance(control, Quantity)
        self.parent_platform.save_action_to_platform(control)

    def get_applied_control(self) -> Quantity:
        return self.parent_platform.get_applied_action()


PluginLibrary.AddClassToGroup(
    GymnasiumMainController, "Controller_Gymnasium", {"simulator": GymnasiumSimulator, "platform_type": GymnasiumAvailablePlatformTypeMain}
)



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
SimpleMultiPlatform Agent Class
"""
import collections
import typing
from functools import cached_property
from graphlib import TopologicalSorter

import gymnasium
from ray.rllib.utils.spaces.repeated import Repeated
from tree import map_structure

from corl.agents.base_agent import BaseAgent, BaseAgentParser
from corl.glues.base_glue import BaseAgentGlue, TrainingExportBehavior
from corl.libraries.env_space_util import EnvSpaceUtil
from corl.libraries.property import DictProp, RepeatedProp
from corl.libraries.space_transformations import SpaceTransformationBase
from corl.libraries.units import corl_get_ureg


class SimpleMultiPlatformParser(BaseAgentParser):
    """
    SimpleMultiPlatformParser is currently the same as BaseAgentParser
    """


class SimpleMultiPlatform(BaseAgent):
    """
    SimpleMultiPlatform is a platform that uses the same configuration as BaseAgent
    but will control multiple platforms with the same parts, glues, rewards, and dones

    This agent does not implement behaviors for glues that do not associate with a platform
    """

    def __init__(self, **kwargs) -> None:
        self.config: SimpleMultiPlatformParser = self.get_validator()(**kwargs)
        self.agent_glue_dict: dict[str, dict[str, BaseAgentGlue]] = {}  # type: ignore
        self.agent_reward_list: list = []
        self.agent_glue_order: dict[str, tuple] = {}  # type: ignore

        self._action_space_transformation: SpaceTransformationBase | None = None
        self._obs_space_transformation: SpaceTransformationBase | None = None

        # Sample parameter provider
        # This RNG only used here.  Normal use uses the one from the environment.
        rng, _ = gymnasium.utils.seeding.np_random(0)
        self.fill_parameters(rng=rng, default_parameters=True)
        self.inoperable_platforms = {platform_name: False for platform_name in self.config.platform_names}

        self.observation_space: gymnasium.spaces.Dict
        self.action_space: gymnasium.spaces.Dict

    @staticmethod
    def get_validator() -> type[SimpleMultiPlatformParser]:
        return SimpleMultiPlatformParser

    def make_glues(self, platforms, agent_id: str, env_ref_stores: list[dict[str, typing.Any]]) -> None:
        """
        Creates agent glue functors from agent configuration.

        Parameters
        ----------
        platforms: The platform_name to platform instances mapping associated with the glue functors.
        agent_id: The id of the agent associated with the glue functors.
        env_ref_stores: Reference stores for items managed by the environment

        Returns
        -------
        None
        """
        self.agent_glue_dict.clear()
        for config_glues, platform_name in zip(self.config.glues, self.config.platform_names):
            platform = platforms[platform_name]
            platform_glues = {}
            sort_order = {}
            for glue_dict in config_glues:
                if self.config.normalization_master_switch == "ALL_OFF":
                    glue_dict.config["normalization"] = {"enabled": False}
                created_glue = glue_dict.create_functor_object(
                    platform=platform,
                    agent_name=agent_id,
                    param_sources=[self.local_variable_store.get("glues", {})],
                    ref_sources=[self.local_variable_store.get("reference_store", {}), self.config.reference_store, *env_ref_stores],
                )
                glue_name = created_glue.get_unique_name()
                if glue_name is None:
                    raise RuntimeError(f"No glue name for {created_glue}")

                if glue_name in platform_glues:
                    raise RuntimeError(f"The {glue_name} glue has a unique name, but it already exists")
                # find dependent nodes for the topological search
                # this will get overwritten each platform loop, but it doesn't matter for this agent class
                dependent_glues = {extractor.fields[0] for extractor in created_glue.config.extractors.values()}
                sort_order[glue_name] = dependent_glues

                # add the glue to the agent glue dict
                # self._agent_glue_obs_export_behavior[glue_name] = glue.training_obs_behavior
                platform_glues[glue_name] = created_glue

            if len(self.config.platform_names) == 1:
                self.agent_glue_dict = platform_glues
            else:
                self.agent_glue_dict[platform_name] = platform_glues

            self.agent_glue_order[platform_name] = tuple(TopologicalSorter(sort_order).static_order())

    def _create_space(self, space_getter: typing.Callable | None = None):
        """
        Creates a gymnasium dict space from the agent's glues.

        Parameters
        ----------
        space_getter (optional): A function that takes a glue_obj: BaseAgentGlue and returns a space.
            Default is space_getter=lambda glue_obj: glue_obj.observation_space.

        Returns
        -------
        gymnasium.spaces.dict.Dict()
            A gymnasium dict space composed of all the spaces returned by applying the space_getter to the agent glues.
        """
        if space_getter is None:

            def default_getter(glue_obj):
                return glue_obj.observation_space

            space_getter = default_getter

        return_space = gymnasium.spaces.dict.Dict()
        # loop over all glue name and  glue_obj pairs
        for platform_name, glue_dict in self.agent_glue_dict.items():
            platform_space = gymnasium.spaces.dict.Dict()
            for glue_name in self.agent_glue_order[platform_name]:
                # call our space getter to pick which space we want,
                # for example: action_space, observation_space, normalized_action_space, normalized_observation_space
                space_def = space_getter(glue_dict[glue_name])
                # if the space is None don't add it
                if space_def:
                    platform_space[glue_name] = space_def
            if len(platform_space) > 0:
                return_space[platform_name] = platform_space
        return return_space if len(return_space) > 0 else None

    def _create_space_dict(self, space_getter: typing.Callable | None = None) -> dict[str, dict[str, DictProp]]:
        """
        Creates a gymnasium dict space from the agent's glues.

        Parameters
        ----------
        space_getter (optional): A function that takes a glue_obj: BaseAgentGlue and returns a space.
            Default is space_getter=lambda glue_obj: glue_obj.observation_space.

        Returns
        -------
        gymnasium.spaces.dict.Dict()
            A gymnasium dict space composed of all the spaces returned by applying the space_getter to the agent glues.
        """
        if space_getter is None:

            def default_getter(glue_obj):
                return glue_obj.observation_space

            space_getter = default_getter

        return_space = {}
        # loop over all glue name and  glue_obj pairs
        for platform_name, glue_dict in self.agent_glue_dict.items():
            platform_space = {}
            for glue_name in self.agent_glue_order[platform_name]:
                # call our space getter to pick which space we want,
                # for example: action_space, observation_space, normalized_action_space, normalized_observation_space
                space_def = space_getter(glue_dict[glue_name])
                # if the space is None don't add it
                if space_def:
                    platform_space[glue_name] = space_def
            if len(platform_space) > 0:
                return_space[platform_name] = platform_space
        if len(return_space) > 0:
            return return_space

        raise RuntimeError("Failed to create space dict")

    def get_observations(self) -> collections.OrderedDict:
        """
        Gets combined observation from agent glues.

        Returns
        -------
        OrderedDict
            A dictionary of glue observations in the form {glue_name: glue_observation}
        """
        return_observation: collections.OrderedDict = collections.OrderedDict()
        for platform_name, glue_dict in self.agent_glue_dict.items():
            if self.inoperable_platforms[platform_name] and glue_dict:
                return_observation[platform_name] = []
                continue
            platform_obs: collections.OrderedDict = collections.OrderedDict()
            for glue_name in self.agent_glue_order[platform_name]:
                glue_object = glue_dict[glue_name]
                if glue_obs := glue_object.get_observation(
                    platform_obs,
                    self.observation_space[platform_name],
                    self.observation_units[platform_name],
                ):
                    platform_obs[glue_name] = glue_obs
                    if glue_object.config.clip_to_space and glue_object.observation_space:
                        platform_obs[glue_name] = EnvSpaceUtil.clip_space_sample_to_space(glue_obs, glue_object.observation_space)
            if len(platform_obs) > 0:
                return_observation[platform_name] = [platform_obs]
        return return_observation

    def create_training_observations(self, observations: collections.OrderedDict):
        """
        Base class representing a trainable agent in an environment.
        """

        training_obs: dict[str, list[dict[str, typing.Any]]] = collections.OrderedDict()
        for platform_name, obs_list in observations.items():
            if len(obs_list) == 0:
                training_obs[platform_name] = []
                continue
            obs_dict = obs_list[0]
            platform_obs = collections.OrderedDict()
            for obs_name, obs in obs_dict.items():
                glue_obj = self.agent_glue_dict[platform_name][obs_name]
                if glue_obj is not None and glue_obj.config.training_export_behavior == TrainingExportBehavior.INCLUDE:
                    platform_obs[obs_name] = glue_obj.normalize_observation(obs)
            training_obs[platform_name] = [platform_obs]

        if self._obs_space_transformation:
            return self._obs_space_transformation.convert_sample(sample=training_obs)
        return training_obs

    def apply_action(self, action: dict[typing.Any, typing.Any] | typing.Any, last_obs: typing.OrderedDict):
        """
        Applies actions to agent.

        Parameters
        ----------
        action (optional): Either a dictionary of actions to be applied to agent, or
            a flattened action.

        Returns
        -------
        None
        """

        if self._action_space_transformation:
            action = self._action_space_transformation.convert_transformed_sample(action)

        raw_action_dict: dict[str, dict[str, typing.Any]] = collections.OrderedDict()
        for platform_name, platform_action in action.items():
            raw_action_dict[platform_name] = {}
            for glue_name, normalized_action in platform_action.items():
                glue_object = self.agent_glue_dict[platform_name][glue_name]
                raw_action = glue_object.unnormalize_action(normalized_action)
                raw_action = map_structure(
                    lambda x, y: corl_get_ureg().Quantity(x, y), raw_action, self.action_units[platform_name][glue_name]
                )
                raw_action_dict[platform_name][glue_name] = raw_action
                if not self.inoperable_platforms[platform_name]:
                    glue_object.apply_action(
                        raw_action,
                        last_obs,
                        self.action_space[platform_name],
                        self.observation_space[platform_name],
                        self.observation_units[platform_name],
                    )
        return raw_action_dict

    def get_info_dict(self) -> collections.OrderedDict:
        """
        Gets combined observation from agent glues.

        Returns
        -------
        OrderedDict
            A dictionary of glue observations in the form {glue_name: glue_observation}
        """
        return_info: collections.OrderedDict = collections.OrderedDict()
        for platform_name, platform_glues in self.agent_glue_dict.items():
            return_info[platform_name] = {}
            for glue_name in self.agent_glue_order[platform_name]:
                if glue_info := platform_glues[glue_name].get_info_dict():
                    return_info[platform_name][glue_name] = glue_info
        return return_info

    def set_removed(self, removed_state: dict[str, bool]):
        """
        Set the agent_removed flag for the agent's glues.

        Parameters
        ----------
        - removed_state (typing.Dict[str, bool]): The the inoperable status of the platform an agent class is controlling.
            True: the platform is no longer operable
            False: the platform is still operable

        Returns
        -------
        None
        """
        for platform, platform_glue_dict in filter(lambda x: removed_state.get(x[0], False), self.agent_glue_dict.items()):
            self.inoperable_platforms[platform] = True
            for glue in platform_glue_dict.values():
                glue.set_agent_removed(True)

    @property
    def trainable(self) -> bool:
        """
        Flag denoting if agent is trainable.

        Returns
        -------
        bool:
            True
        """
        return True

    @cached_property
    def _observation_space(self):
        tmp = self._create_space(space_getter=lambda glue_obj: glue_obj.observation_space)
        return gymnasium.spaces.Dict(
            {platform_name: Repeated(platform_space, max_len=1) for platform_name, platform_space in tmp.spaces.items()}
        )

    @cached_property
    def action_prop(self):
        """
        Returns the action prop that defines this agents actions
        """
        tmp = self._create_space_dict(space_getter=lambda glue_obj: glue_obj.action_prop)
        assert tmp is not None
        return DictProp(spaces={platform_name: DictProp(spaces=tmp[platform_name]) for platform_name in self.agent_glue_dict})

    @cached_property
    def observation_prop(self):
        """
        Returns the action prop that defines this agents actions
        """
        tmp = self._create_space_dict(space_getter=lambda glue_obj: glue_obj.observation_prop)
        assert tmp is not None
        return DictProp(
            spaces={platform_name: RepeatedProp(child_space=tmp[platform_name], max_len=1) for platform_name in self.agent_glue_dict}
        )

    @cached_property
    def normalized_observation_space(self):
        """
        Returns agents normalized observation space
        """
        tmp = self._create_space(
            space_getter=lambda glue_obj: glue_obj.normalized_observation_space
            if glue_obj.config.training_export_behavior == TrainingExportBehavior.INCLUDE
            else None
        )
        tmp = gymnasium.spaces.Dict(
            {platform_name: Repeated(platform_space, max_len=1) for platform_name, platform_space in tmp.spaces.items()}
        )
        if self.config.policy_observation_transformation:
            self._obs_space_transformation = self.config.policy_observation_transformation.transformation(
                tmp, self.config.policy_observation_transformation.config
            )
            if self._obs_space_transformation:
                return self._obs_space_transformation.output_space
        return tmp



"""
---------------------------------------------------------------------------


Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Policy Network Visualization
"""
import streamlit as st

from corl.evaluation.recording.folder import EpisodeArtifactLoggingCallbackLoader, FolderRecord
from corl.visualization.streamlit_app.default_layout import DefaultStreamlitPage
from corl.visualization.streamlit_app.eval_while_train_layout import StreamlitPageEvaluationWhileTraining

if __name__ == "__main__":
    st.set_page_config(layout="wide")

    # #########################################################################
    # Policy Network Visualization Tool.
    # #########################################################################
    st.markdown("# Policy Network Visualization")
    st.warning(
        """This page may cause the app to crash.
        Under the hood, this initializes the simulation environment, loads in your trained policies
        and initializes Ray. When the simulation environment is large, it seems to cause the app
        to either freeze or crash.
        """
    )
    with st.form("eval-selection"):
        corl_eval = "CoRL Custom Evaluation"
        eval_while_train = "Evaluation Generated while Training"

        selected_output = st.radio(label="Select Evaluation Type", options=(corl_eval, eval_while_train), horizontal=True)
        submitted = st.form_submit_button("Done")

    if submitted:
        if selected_output == corl_eval:
            st.session_state.policy_network_app = DefaultStreamlitPage(file_loader=FolderRecord)
        else:
            st.session_state.policy_network_app = StreamlitPageEvaluationWhileTraining(file_loader=EpisodeArtifactLoggingCallbackLoader)

    if hasattr(st.session_state, "policy_network_app"):
        st.session_state.policy_network_app.display_directory_picker()
        if st.button("Select Run") or hasattr(st.session_state, "selected_run"):
            st.session_state.selected_run = True
            st.session_state.policy_network_app.display_policy_network_viz_section()



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
Module contains a storage interface which can be extended to implement various kinds of storage for the evaluation framework.
"""

from abc import ABC, abstractmethod


class Storage(ABC):
    """
    Abstract class that defines an interface by which various kinds of storage can be utilized
    """

    def __init__(self) -> None:
        ...

    @abstractmethod
    def load_artifacts_location(self, config: dict[str, str]):
        """
        Load information about the locations of various artifacts
        """

    @abstractmethod
    def store(self):
        """
        Logic to perform the storage of data.
        """



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""
import dataclasses
import typing

from corl.evaluation.metrics.metric import TerminalMetric
from corl.libraries.units import Quantity


@dataclasses.dataclass
class Real(TerminalMetric):
    """Terminal Metric to represent a real number"""

    value: float

    def __add__(self, rhs: typing.Any):
        if isinstance(rhs, Real):
            return Real(self.value + rhs.value)
        raise RuntimeError(f"No logic on how to handle {type(rhs)}")

    def __truediv__(self, rhs: typing.Any):
        lhs = self.value.value if isinstance(self.value, Real) else self.value

        if isinstance(rhs, (Real)):
            rhs = rhs.value
        elif isinstance(rhs, Quantity):
            rhs = rhs.m

        quotient = float(lhs) / float(rhs)

        return Real(quotient)

    def __lt__(self, rhs: typing.Any):
        if isinstance(rhs, Real):
            return self.value < rhs.value
        if isinstance(rhs, int):
            return self.value < typing.cast(float, rhs)
        if isinstance(rhs, float):
            return self.value < rhs
        raise RuntimeError(f"No logic on how to handle {type(rhs)}")

    def __gt__(self, rhs: typing.Any):
        if isinstance(rhs, Real):
            return self.value > rhs.value
        if isinstance(rhs, int):
            return self.value > typing.cast(float, rhs)
        if isinstance(rhs, float):
            return self.value > rhs
        raise RuntimeError(f"No logic on how to handle {type(rhs)}")



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
"""

from corl.evaluation.metrics.generator import MetricGeneratorAggregator
from corl.evaluation.metrics.metric import Metric
from corl.evaluation.metrics.types.terminals.rate import Rate
from corl.evaluation.util.condition import Condition


class CriteriaRate(MetricGeneratorAggregator):
    """Aggregate a metric by determining the rate which a criteria is established"""

    condition: Condition

    def __init__(self, **data):
        super().__init__(**data)
        self.condition = Condition(**data["condition"])

    def generate_metric(self, params: list[Metric] | Metric, **kwargs) -> Metric:
        """Generate the metric

        Arguments:
            metrics {typing.Union[typing.List[Metric], Metric]} -- Already computed metrics

        Returns:
            Metric -- Computed Metric
        """

        if isinstance(params, Metric):
            raise RuntimeError("Metrics given to Criteria Rate must be a list (that is an aggregation of lower scope)")

        occurrence = sum(bool(self.condition.func(item)) for item in params)
        return Rate(occurrence, len(params))



"""
---------------------------------------------------------------------------
Air Force Research Laboratory (AFRL) Autonomous Capabilities Team (ACT3)
Reinforcement Learning (RL) Core.

This is a US Government Work not subject to copyright protection in the US.

The use, dissemination or disclosure of data in this file is subject to
limitation or restriction. See accompanying README and LICENSE for details.
---------------------------------------------------------------------------
This module defines the TestCaseStrategy for running evaluation with the default EPPs defined in the task configs.

Author: John McCarroll
"""


from corl.evaluation.runners.section_factories.test_cases.test_case_manager import (
    TestCaseIndex,
    TestCaseStrategy,
    TestCaseStrategyValidator,
)


class DefaultStrategyValidator(TestCaseStrategyValidator):
    """
    A validator for DefaultStrategy class
    """

    num_test_cases: int = 1
    epp_wrapper_path: str = "corl.episode_parameter_providers.incremental.IncrementalParameterProviderWrapper"


class DefaultStrategy(TestCaseStrategy):
    """
    The concrete implementation of a TestCaseStrategy for a task's default ParameterProviders.
    """

    def __init__(self, **kwargs) -> None:
        self.config: DefaultStrategyValidator
        self.test_cases: list = []
        super().__init__(**kwargs)

    @staticmethod
    def get_validator() -> type[DefaultStrategyValidator]:
        """
        Method to return validator for DefaultStrategyValidator
        """
        return DefaultStrategyValidator

    def update_rllib_config(self, config: dict):
        """
        Method responsible for mutating rllib config, overwriting the task's default EPPs with TabularParameterProviders.
        """
        # wrap EPPs w IncrementalParameterProviderWrapper

        # retrieve epp config for environment's params
        # pass epp config to wrapper
        epp_env_type = config["env_config"]["episode_parameter_provider"]["type"]
        wrapper_config = {
            "num_test_cases": self.config.num_test_cases,
            "type": epp_env_type,
        }
        # add config, if available
        epp_env_config = (
            config["env_config"]["episode_parameter_provider"]["config"]
            if "config" in config["env_config"]["episode_parameter_provider"]
            else None
        )
        if epp_env_config:
            wrapper_config["config"] = epp_env_config

        config["env_config"]["episode_parameter_provider"] = {"type": self.config.epp_wrapper_path, "config": wrapper_config}

        # Wrap the agents EPPs w Incremental EPP
        for value in config["env_config"]["agents"].values():
            # retrieve epp config for agent's params
            epp_agent_type = value.class_config.config["episode_parameter_provider"]["type"]
            wrapper_config = {
                "num_test_cases": self.config.num_test_cases,
                "type": epp_agent_type,
            }
            # add config, if aailable
            epp_agent_config = (
                value.class_config.config["episode_parameter_provider"]["config"]
                if "config" in value.class_config.config["episode_parameter_provider"]
                else None
            )
            if epp_agent_config:
                wrapper_config["config"] = epp_env_config

            value.class_config.config["episode_parameter_provider"] = {"type": self.config.epp_wrapper_path, "config": wrapper_config}

        return config

    def get_num_test_cases(self) -> int:
        return self.config.num_test_cases

    def get_test_cases(self):
        """
        Method responsible for retrieving test cases planned for evaluation.
        Using default strategy does not plan test cases by a collection of parameter values.
        """
        return self.test_cases

    def get_test_case_index(self, episode_id: int) -> TestCaseIndex:  # noqa: PLR6301
        return TestCaseIndex(episode_id)
